'''
My cross trainer exercise data, arbitrated from different sources (mainly, Endomondo and manual text notes)

This is probably too specific to my needs, so later I will move it away to a personal 'layer'.
For now it's worth keeping it here as an example and perhaps utility functions might be useful for other HPI modules.
'''

from datetime import datetime, timedelta
from typing import Optional

from ...core.pandas import DataFrameT, check_dataframe as cdf
from ...core.orgmode import collect, Table, parse_org_datetime, TypedTable

from my.config import exercise as config


import pytz
# FIXME how to attach it properly?
tz = pytz.timezone('Europe/London')

def tzify(d: datetime) -> datetime:
    assert d.tzinfo is None, d
    return tz.localize(d)


# todo predataframe?? entries??
def cross_trainer_data():
    # FIXME some manual entries in python
    # I guess just convert them to org
    import orgparse
    # todo should use all org notes and just query from them?
    wlog = orgparse.load(config.workout_log)

    [table] = collect(
        wlog,
        lambda n: [] if n.heading != 'Cross training' else [x for x in n.body_rich if isinstance(x, Table)]
    )
    cross_table = TypedTable(table)

    def maybe(f):
        def parse(s):
            if len(s) == 0:
                return None
            return f(s)
        return parse

    def parse_mm_ss(x: str) -> timedelta:
        hs, ms = x.split(':')
        return timedelta(seconds=int(hs) * 60 + int(ms))

    # todo eh. not sure if there is a way of getting around writing code...
    # I guess would be nice to have a means of specifying type in the column? maybe multirow column names??
    # need to look up org-mode standard..
    mappers = {
        'duration': lambda s: parse_mm_ss(s),
        'date'    : lambda s: tzify(parse_org_datetime(s)),
        'comment' : str,
    }
    for row in cross_table.as_dicts:
        # todo make more defensive, fallback on nan for individual fields??
        try:
            d = {}
            for k, v in row.items():
                # todo have something smarter... e.g. allow pandas to infer the type??
                mapper = mappers.get(k, maybe(float))
                d[k] = mapper(v) # type: ignore[operator]
            yield d
        except Exception as e:
            # todo add parsing context
            yield {'error': str(e)}

    # todo hmm, converting an org table directly to pandas kinda makes sense?
    # could have a '.dataframe' method in orgparse, optional dependency


@cdf
def cross_trainer_manual_dataframe() -> DataFrameT:
    '''
    Only manual org-mode entries
    '''
    import pandas as pd
    df = pd.DataFrame(cross_trainer_data())
    return df

# this should be enough?..
_DELTA = timedelta(hours=10)

# todo check error handling by introducing typos (e.g. especially dates) in org-mode
@cdf
def dataframe() -> DataFrameT:
    '''
    Attaches manually logged data (which Endomondo can't capture) and attaches it to Endomondo
    '''
    import pandas as pd

    from ...endomondo import dataframe as EDF
    edf = EDF()
    edf = edf[edf['sport'].str.contains('Cross training')]

    mdf = cross_trainer_manual_dataframe()
    # TODO shit. need to always remember to split errors???
    # on the other hand, dfs are always untyped. so it's not too bad??
    # now for each manual entry, find a 'close enough' endomondo entry
    # ideally it's a 1-1 (or 0-1) relationship, but there might be errors
    rows = []
    idxs = [] # type: ignore[var-annotated]
    NO_ENDOMONDO = 'no endomondo matches'
    for i, row in mdf.iterrows():
        rd = row.to_dict()
        mdate = row['date']
        if pd.isna(mdate):
            # todo error handling got to be easier. seriously, mypy friendly dataframes would be amazing
            idxs.append(None)
            rows.append(rd) # presumably has an error set
            continue

        idx: Optional[int]
        close = edf[edf['start_time'].apply(lambda t: pd_date_diff(t, mdate)).abs() < _DELTA]
        if len(close) == 0:
            idx = None
            d = {
                **rd,
                'error': NO_ENDOMONDO,
            }
        elif len(close) > 1:
            idx = None
            d = {
                **rd,
                'error': f'one manual, many endomondo: {close}',
            }
        else:
            idx = close.index[0]
            d = rd

            if idx in idxs:
                # todo might be a good idea to remove the original match as well?
                idx = None
                d = {
                    **rd,
                    'error': 'one endomondo, many manual',
                }
        idxs.append(idx)
        rows.append(d)
    mdf = pd.DataFrame(rows, index=idxs)

    # todo careful about 'how'? we need it to preserve the errors
    # maybe pd.merge is better suited for this??
    df = edf.join(mdf, how='outer', rsuffix='_manual')
    # todo reindex? so we don't have Nan leftovers

    # todo set date anyway? maybe just squeeze into the index??
    noendo = df['error'] == NO_ENDOMONDO
    # meh. otherwise the column type ends up object
    tz = df[noendo]['start_time'].dtype.tz
    df.loc[noendo, 'start_time'    ] = df[noendo]['date'           ].dt.tz_convert(tz)
    df.loc[noendo, 'duration'      ] = df[noendo]['duration_manual']
    df.loc[noendo, 'heart_rate_avg'] = df[noendo]['hr_avg'         ]

    # todo set sport?? set source?
    return df
# TODO arbitrate kcal, duration, avg hr
# compare power and hr? add 'quality' function??
# TODO wtf?? where is speed coming from??


from ...core import stat, Stats
def stats() -> Stats:
    return stat(cross_trainer_data)


def compare_manual() -> None:
    df = dataframe()
    df = df.set_index('start_time')

    df = df[[
        'kcal'    , 'kcal_manual',
        'duration', 'duration_manual',
    ]].dropna()
    print(df.to_string())


def pd_date_diff(a, b) -> timedelta:
    # ugh. pandas complains when we subtract timestamps in different timezones
    assert a.tzinfo is not None, a
    assert b.tzinfo is not None, b
    return a.to_pydatetime() - b.to_pydatetime()



"""
[[https://play.google.com/store/apps/details?id=io.github.hidroh.materialistic][Materialistic]] app for Hackernews
"""
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Iterator, NamedTuple, Sequence

from more_itertools import unique_everseen

from my.core import get_files, datetime_aware, make_logger
from my.core.sqlite import sqlite_connection

from my.config import materialistic as config  # todo migrate config to my.hackernews.materialistic

from .common import hackernews_link


logger = make_logger(__name__)


def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


Row = Dict[str, Any]


class Saved(NamedTuple):
    row: Row

    # NOTE: seems like it's the time item was saved (not created originally??)
    # https://github.com/hidroh/materialistic/blob/b631d5111b7487d2328f463bd95e8507c74c3566/app/src/main/java/io/github/hidroh/materialistic/data/MaterialisticDatabase.java#L224
    # but not 100% sure.
    @property
    def when(self) -> datetime_aware:
        ts = int(self.row['time']) / 1000
        return datetime.fromtimestamp(ts, tz=timezone.utc)

    @property
    def uid(self) -> str:
        return self.row['itemid']

    @property
    def url(self) -> str:
        return self.row['url']

    @property
    def title(self) -> str:
        return self.row['title']

    @property
    def hackernews_link(self) -> str:
        return hackernews_link(self.uid)


def _all_raw() -> Iterator[Row]:
    paths = inputs()
    total = len(paths)
    width = len(str(total))
    for idx, path in enumerate(paths):
        logger.info(f'processing [{idx:>{width}}/{total:>{width}}] {path}')
        with sqlite_connection(path, immutable=True, row_factory='dict') as conn:
            yield from conn.execute('SELECT * FROM saved ORDER BY time')


def raw() -> Iterator[Row]:
    yield from unique_everseen(_all_raw(), key=lambda r: r['itemid'])


def saves() -> Iterator[Saved]:
    yield from map(Saved, raw())



import os
from pathlib import Path
import shutil
import tempfile
from typing import TYPE_CHECKING
import zipfile

from ..common import get_files
from ..compat import windows
from ..kompress import CPath, ZipPath

import pytest


# hack to replace all /tmp with 'real' tmp dir
# not ideal, but makes tests more concise
# TODO get rid of this, it's super confusing..
def _get_files(x, *args, **kwargs):
    from ..common import get_files as get_files_orig

    def repl(x):
        if isinstance(x, str):
            return x.replace('/tmp', TMP)
        elif isinstance(x, Path):
            assert x.parts[:2] == (os.sep, 'tmp')  # meh
            return Path(TMP) / Path(*x.parts[2:])
        else:
            # iterable?
            return [repl(i) for i in x]

    x = repl(x)
    res = get_files_orig(x, *args, **kwargs)
    return tuple(type(i)(str(i).replace(TMP, '/tmp')) for i in res)  # hack back for asserts..


get_files_orig = get_files
if not TYPE_CHECKING:
    get_files = _get_files


def test_single_file() -> None:
    '''
    Regular file path is just returned as is.
    '''

    "Exception if it doesn't exist"
    with pytest.raises(Exception):
        get_files('/tmp/hpi_test/file.ext')

    create('/tmp/hpi_test/file.ext')

    '''
    Couple of things:
    1. Return type is a tuple, it's friendlier for hashing/caching
    2. It always return pathlib.Path instead of plain strings
    '''
    assert get_files('/tmp/hpi_test/file.ext') == (Path('/tmp/hpi_test/file.ext'),)

    "if the path starts with ~, we expand it"
    if not windows:  # windows doesn't have bashrc.. ugh
        assert get_files('~/.bashrc') == (Path('~').expanduser() / '.bashrc',)


def test_multiple_files() -> None:
    '''
    If you pass a directory/multiple directories, it flattens the contents
    '''
    create('/tmp/hpi_test/dir1/')
    create('/tmp/hpi_test/dir1/zzz')
    create('/tmp/hpi_test/dir1/yyy')
    # create('/tmp/hpi_test/dir1/whatever/') # TODO not sure about this... should really allow extra dirs
    create('/tmp/hpi_test/dir2/')
    create('/tmp/hpi_test/dir2/mmm')
    create('/tmp/hpi_test/dir2/nnn')
    create('/tmp/hpi_test/dir3/')
    create('/tmp/hpi_test/dir3/ttt')

    # fmt: off
    assert get_files([
        Path('/tmp/hpi_test/dir3'), # it takes in Path as well as str
        '/tmp/hpi_test/dir1',
    ]) == (
        # the paths are always returned in sorted order (unless you pass sort=False)
        Path('/tmp/hpi_test/dir1/yyy'),
        Path('/tmp/hpi_test/dir1/zzz'),
        Path('/tmp/hpi_test/dir3/ttt'),
    )
    # fmt: on


def test_explicit_glob() -> None:
    '''
    You can pass a glob to restrict the extensions
    '''

    create('/tmp/hpi_test/file_3.gz')
    create('/tmp/hpi_test/file_2.gz')
    create('/tmp/hpi_test/ignoreme')
    create('/tmp/hpi_test/file.gz')

    # todo walrus operator would be great here...
    expected = (
        Path('/tmp/hpi_test/file_2.gz'),
        Path('/tmp/hpi_test/file_3.gz'),
    )
    assert get_files('/tmp/hpi_test', 'file_*.gz') == expected

    "named argument should work too"
    assert get_files('/tmp/hpi_test', glob='file_*.gz') == expected


def test_implicit_glob() -> None:
    '''
    Asterisc in the path results in globing too.
    '''
    # todo hopefully that makes sense? dunno why would anyone actually rely on asteriscs in names..
    # this is very convenient in configs, so people don't have to use some special types

    create('/tmp/hpi_test/123/')
    create('/tmp/hpi_test/123/dummy')
    create('/tmp/hpi_test/123/file.gz')
    create('/tmp/hpi_test/456/')
    create('/tmp/hpi_test/456/dummy')
    create('/tmp/hpi_test/456/file.gz')

    assert get_files(['/tmp/hpi_test/*/*.gz']) == (
        Path('/tmp/hpi_test/123/file.gz'),
        Path('/tmp/hpi_test/456/file.gz'),
    )


def test_no_files() -> None:
    '''
    Test for empty matches. They work, but should result in warning
    '''
    assert get_files('') == ()

    # todo test these for warnings?
    assert get_files([]) == ()
    assert get_files('bad*glob') == ()


def test_compressed(tmp_path: Path) -> None:
    file1 = tmp_path / 'file_1.zstd'
    file2 = tmp_path / 'file_2.zip'
    file3 = tmp_path / 'file_3.csv'

    file1.touch()
    with zipfile.ZipFile(file2, 'w') as zf:
        zf.writestr('path/in/archive', 'data in zip')
    file3.touch()

    results = get_files_orig(tmp_path)
    [res1, res2, res3] = results
    assert isinstance(res1, CPath)
    assert isinstance(res2, ZipPath)  # NOTE this didn't work on vendorized kompress, but it's fine, was never used?
    assert not isinstance(res3, CPath)

    results = get_files_orig(
        [CPath(file1), ZipPath(file2), file3],
        # sorting a mixture of ZipPath/Path was broken in old kompress
        # it almost never happened though (usually it's only a bunch of ZipPath, so not a huge issue)
        sort=False,
    )
    [res1, res2, res3] = results
    assert isinstance(res1, CPath)
    assert isinstance(res2, ZipPath)
    assert not isinstance(res3, CPath)


# TODO not sure if should uniquify if the filenames end up same?
# TODO not sure about the symlinks? and hidden files?

TMP = tempfile.gettempdir()
test_path = Path(TMP) / 'hpi_test'


def setup():
    teardown()
    test_path.mkdir()


def teardown():
    if test_path.is_dir():
        shutil.rmtree(test_path)


def create(f: str) -> None:
    # in test body easier to use /tmp regardless the OS...
    f = f.replace('/tmp', TMP)
    if f.endswith('/'):
        Path(f).mkdir()
    else:
        Path(f).touch()



'''
Foursquare/Swarm checkins
'''

from datetime import datetime, timezone, timedelta
from itertools import chain
from pathlib import Path
import json

# TODO pytz for timezone???

from .core.common import get_files, LazyLogger
from my.config import foursquare as config


logger = LazyLogger(__name__)


def inputs():
    return get_files(config.export_path)


class Checkin:
    def __init__(self, j) -> None:
        self.j = j

    @property
    def summary(self) -> str:
        name = self.j.get('venue', {}).get('name', 'NO_NAME')
        return "checked into " + name + " " + self.j.get('shout', "") # TODO should should be bold...

    @property
    def dt(self) -> datetime:
        created = self.j['createdAt']  # this is local time
        offset = self.j['timeZoneOffset']
        tz = timezone(timedelta(minutes=offset))
        # a bit meh, but seems to work..
        # TODO localize??
        return datetime.fromtimestamp(created, tz=tz)

    @property
    def cid(self) -> str:
        return self.j['id']

    def __repr__(self):
        return repr(self.j)


class Place:
    def __init__(self, j) -> None:
        self.j = j


# TODO ugh. I'm not backing up lists, apparently...
# def test_places():
#     raise RuntimeError()


# TODO need json type

def get_raw(fname=None):
    if fname is None:
        fname = max(inputs())
    j = json.loads(fname.read_text())
    assert isinstance(j, list)

    for chunk in j:
        del chunk['meta']
        del chunk['notifications']
    assert chunk.keys() == {'response'}
    assert chunk['response'].keys() == {'checkins'}

    return chain.from_iterable(x['response']['checkins']['items'] for x in j)


# TODO not sure how to make it generic..
def get_checkins(*args, **kwargs):
    everything = get_raw(*args, **kwargs)
    checkins = sorted([Checkin(i) for i in everything], key=lambda c: c.dt)
    return checkins


# TODO do I need this??
def get_cid_map(bfile: str):
    raw = get_raw(bfile)
    return {i['id']: i for i in raw}


def print_checkins():
    print(get_checkins())


def stats():
    from more_itertools import ilen
    return {
        'checkins': ilen(get_checkins()),
    }



from typing import Protocol

from my.core import datetime_aware, Json


def hackernews_link(id: str) -> str:
    return f'https://news.ycombinator.com/item?id={id}'


class SavedBase(Protocol):
    @property
    def when(self) -> datetime_aware: ...
    @property
    def uid(self) -> str: ...
    @property
    def url(self) -> str: ...
    @property
    def title(self) -> str: ...
    @property
    def hackernews_link(self) -> str: ...



from pathlib import Path


def test_dynamic_configuration(notes: Path) -> None:
    import pytz
    from types import SimpleNamespace as NS

    from my.core.cfg import tmp_config
    with tmp_config() as C:
        C.orgmode = NS(paths=[notes])
        # TODO ugh. this belongs to tz provider or global config or something
        C.weight  = NS(default_timezone=pytz.timezone('Europe/London'))

        from my.body.weight import from_orgmode
        weights = [0.0 if isinstance(x, Exception) else x.value for x in from_orgmode()]

    assert weights == [
        0.0,
        62.0,
        0.0,
        61.0,
        62.0,
        0.0,
    ]

import pytest


def test_environment_variable(tmp_path: Path) -> None:
    cfg_dir  = tmp_path / 'my'
    cfg_file = cfg_dir / 'config.py'
    cfg_dir.mkdir()
    cfg_file.write_text('''
class feedly:
    pass
class just_for_test:
    pass
''')

    import os
    oenv = dict(os.environ)
    try:
        os.environ['MY_CONFIG'] = str(tmp_path)
        # should not raise at least
        import my.rss.feedly

        import my.config as c
        assert hasattr(c, 'just_for_test')
    finally:
        os.environ.clear()
        os.environ.update(oenv)

    import sys
    # TODO wtf??? doesn't work without unlink... is it caching something?
    cfg_file.unlink()
    del sys.modules['my.config'] # meh..

    import my.config as c
    assert not hasattr(c, 'just_for_test')


from dataclasses import dataclass


def test_user_config() -> None:
    from my.core.common import classproperty
    class user_config:
        param1 = 'abacaba'
        # TODO fuck. properties don't work here???
        @classproperty
        def param2(cls) -> int:
            return 456

        extra = 'extra!'

    @dataclass
    class test_config(user_config):
        param1: str
        param2: int # type: ignore[assignment] # TODO need to figure out how to trick mypy for @classproperty
        param3: str = 'default'

    assert test_config.param1 == 'abacaba'
    assert test_config.param2 == 456
    assert test_config.param3 == 'default'
    assert test_config.extra  == 'extra!'

    from my.core.cfg import make_config
    c = make_config(test_config)
    assert c.param1 == 'abacaba'
    assert c.param2 == 456
    assert c.param3 == 'default'
    assert c.extra  == 'extra!'


@pytest.fixture
def notes(tmp_path: Path):
    ndir = tmp_path / 'notes'
    ndir.mkdir()
    logs = ndir / 'logs.org'
    logs.write_text('''
#+TITLE: Stuff I'm logging

* Weight (org-capture) :weight:
** [2020-05-01 Fri 09:00] 62
** 63
    this should be ignored, got no timestamp
** [2020-05-03 Sun 08:00] 61
** [2020-05-04 Mon 10:00] 62
    ''')
    misc = ndir / 'misc.org'
    misc.write_text('''
Some misc stuff

* unrelated note :weight:whatever:
    ''')
    try:
        yield ndir
    finally:
        pass


@pytest.fixture(autouse=True)
def prepare():
    from my.tests.common import reset_modules
    reset_modules()
    yield



from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
import shutil
import sqlite3
from tempfile import TemporaryDirectory

from ..sqlite import sqlite_connect_immutable, sqlite_copy_and_open


def test_sqlite_read_with_wal(tmp_path: Path) -> None:
    db = tmp_path / 'db.sqlite'
    # write a bit
    with sqlite3.connect(str(db)) as conn:
        conn.execute('CREATE TABLE testtable (col)')
        for i in range(5):
            conn.execute('INSERT INTO testtable (col) VALUES (?)', str(i))

    # write more in WAL mode
    with sqlite3.connect(str(db)) as conn_db:
        conn.execute('PRAGMA journal_mode=wal;')
        for i in range(5, 10):
            conn_db.execute('INSERT INTO testtable (col) VALUES (?)', str(i))
        conn_db.execute('COMMIT')

        # make sure it has unflushed stuff in wal
        wals = list(db.parent.glob('*-wal'))
        assert len(wals) == 1

        ## now run the tests in separate process to ensure there is no potential for reusing sqlite connections or something
        with ProcessPoolExecutor(1) as pool:
            # merely using it for ctx manager..
            # fmt: off
            pool.submit(_test_do_copy         , db).result()
            pool.submit(_test_do_immutable    , db).result()
            pool.submit(_test_do_copy_and_open, db).result()
            pool.submit(_test_open_asis       , db).result()
            # fmt: on


def _test_do_copy(db: Path) -> None:
    # from a copy without journal can only read previously committed stuff
    with TemporaryDirectory() as tdir:
        cdb = Path(tdir) / 'dbcopy.sqlite'
        shutil.copy(db, cdb)
        with sqlite3.connect(str(cdb)) as conn_copy:
            assert len(list(conn_copy.execute('SELECT * FROM testtable'))) == 5
        conn_copy.close()


def _test_do_immutable(db: Path) -> None:
    # in readonly mode doesn't touch
    with sqlite_connect_immutable(db) as conn_imm:
        assert len(list(conn_imm.execute('SELECT * FROM testtable'))) == 5
    conn_imm.close()


def _test_do_copy_and_open(db: Path) -> None:
    with sqlite_copy_and_open(db) as conn_mem:
        assert len(list(conn_mem.execute('SELECT * FROM testtable'))) == 10
    conn_mem.close()


def _test_open_asis(db: Path) -> None:
    # NOTE: this also works... but leaves some potential for DB corruption
    with sqlite3.connect(str(db)) as conn_db_2:
        assert len(list(conn_db_2.execute('SELECT * FROM testtable'))) == 10
    conn_db_2.close()



'''
[[https://play.google.com/store/apps/details?id=com.waterbear.taglog][Taplog]] app data
'''
from datetime import datetime
from typing import NamedTuple, Dict, Optional, Iterable

from my.core import get_files, stat, Stats
from my.core.sqlite import sqlite_connection

from my.config import taplog as user_config


class Entry(NamedTuple):
    row: Dict

    @property
    def id(self) -> str:
        return str(self.row['_id'])

    @property
    def number(self) -> Optional[float]:
        ns = self.row['number']
        # TODO ??
        if isinstance(ns, str):
            ns = ns.strip()
            return None if len(ns) == 0 else float(ns)
        else:
            return ns

    @property
    def note(self) -> str:
        return self.row['note']

    @property
    def button(self) -> str:
        return self.row['cat1']

    @property
    def timestamp(self) -> datetime:
        ts = self.row['timestamp']
        # already with timezone apparently
        # TODO not sure if should still localize though? it only kept tz offset, not real tz
        return datetime.fromisoformat(ts)
    # TODO also has gps info!


def entries() -> Iterable[Entry]:
    last = max(get_files(user_config.export_path))
    with sqlite_connection(last, immutable=True, row_factory='dict') as db:
        # todo is it sorted by timestamp?
        for row in db.execute('SELECT * FROM Log'):
            yield Entry(row)


# I guess worth having as top level considering it would be quite common?
def by_button(button: str) -> Iterable[Entry]:
    for e in entries():
        if e.button == button:
            yield e


def stats() -> Stats:
    return stat(entries)



"""
Parses Google Takeout using [[https://github.com/seanbreckenridge/google_takeout_parser][google_takeout_parser]]

See [[https://github.com/seanbreckenridge/google_takeout_parser][google_takeout_parser]] for more information
about how to export and organize your takeouts

If the DISABLE_TAKEOUT_CACHE environment variable is set, this won't cache individual
exports in ~/.cache/google_takeout_parser

The directory set as takeout_path can be unpacked directories, or
zip files of the exports, which are temporarily unpacked while creating
the cachew cache
"""

REQUIRES = ["git+https://github.com/seanbreckenridge/google_takeout_parser"]

from contextlib import ExitStack
import os
from typing import List, Sequence, cast
from pathlib import Path
from my.core import make_config, dataclass
from my.core.common import Stats, LazyLogger, mcachew, get_files, Paths
from my.core.error import ErrorPolicy
from my.core.structure import match_structure

from my.core.time import user_forced
from google_takeout_parser.parse_html.html_time_utils import ABBR_TIMEZONES
ABBR_TIMEZONES.extend(user_forced())

import google_takeout_parser
from google_takeout_parser.path_dispatch import TakeoutParser
from google_takeout_parser.merge import GoogleEventSet, CacheResults

# see https://github.com/seanbreckenridge/dotfiles/blob/master/.config/my/my/config/__init__.py for an example
from my.config import google as user_config


@dataclass
class google(user_config):
    # directory which includes unpacked/zipped takeouts
    takeout_path: Paths

    error_policy: ErrorPolicy = 'yield'

    # experimental flag to use core.kompress.ZipPath
    # instead of unpacking to a tmp dir via match_structure
    _use_zippath: bool = False


config = make_config(google)


logger = LazyLogger(__name__, level="warning")

# patch the takeout parser logger to match the computed loglevel
from google_takeout_parser.log import setup as setup_takeout_logger
setup_takeout_logger(logger.level)


DISABLE_TAKEOUT_CACHE = "DISABLE_TAKEOUT_CACHE" in os.environ


def inputs() -> Sequence[Path]:
    return get_files(config.takeout_path)


try:
    from google_takeout_parser.locales.main import get_paths_for_functions

    EXPECTED = tuple(get_paths_for_functions())

except ImportError:
    EXPECTED = (
        "My Activity",
        "Chrome",
        "Location History",
        "Youtube",
        "YouTube and YouTube Music",
    )


google_takeout_version = str(getattr(google_takeout_parser, '__version__', 'unknown'))

def _cachew_depends_on() -> List[str]:
    exports = sorted([str(p) for p in inputs()])
    # add google takeout parser pip version to hash, so this re-creates on breaking changes
    exports.insert(0, f"google_takeout_version: {google_takeout_version}")
    return exports


# ResultsType is a Union of all of the models in google_takeout_parser
@mcachew(depends_on=_cachew_depends_on, logger=logger, force_file=True)
def events(disable_takeout_cache: bool = DISABLE_TAKEOUT_CACHE) -> CacheResults:
    error_policy = config.error_policy
    count = 0
    emitted = GoogleEventSet()
    # reversed shouldn't really matter? but logic is to use newer
    # takeouts if they're named according to date, since JSON Activity
    # is nicer than HTML Activity
    for path in reversed(inputs()):
        with ExitStack() as exit_stack:
            if config._use_zippath:
                # for later takeouts it's just 'Takeout' dir,
                # but for older (pre 2015) it contains email/date in the subdir name
                results = tuple(cast(Sequence[Path], path.iterdir()))
            else:
                results = exit_stack.enter_context(match_structure(path, expected=EXPECTED, partial=True))
            for m in results:
                # e.g. /home/sean/data/google_takeout/Takeout-1634932457.zip") -> 'Takeout-1634932457'
                # means that zipped takeouts have nice filenames from cachew
                cw_id, _, _ = path.name.rpartition(".")
                # each takeout result is cached as well, in individual databases per-type
                tk = TakeoutParser(m, cachew_identifier=cw_id, error_policy=error_policy)
                # TODO might be nice to pass hpi cache dir?
                for event in tk.parse(cache=not disable_takeout_cache):
                    count += 1
                    if isinstance(event, Exception):
                        if error_policy == 'yield':
                            yield event
                        elif error_policy == 'raise':
                            raise event
                        elif error_policy == 'drop':
                            pass
                        continue
                    if event in emitted:
                        continue
                    emitted.add(event)
                    yield event  # type: ignore[misc]
    logger.debug(
        f"HPI Takeout merge: from a total of {count} events, removed {count - len(emitted)} duplicates"
    )


def stats() -> Stats:
    from my.core import stat

    return {
        **stat(events),
    }



"""
Twitter data (uses [[https://help.twitter.com/en/managing-your-account/how-to-download-your-twitter-archive][official twitter archive export]])
"""


# before this config was named 'twitter', doesn't make too much sense for archive
# todo unify with other code like this, e.g. time.tz.via_location
try:
    from my.config import twitter_archive as user_config
except ImportError as ie:
    if not (ie.name == 'my.config' and 'twitter_archive' in str(ie)):
        # must be caused by something else
        raise ie
    try:
        from my.config import twitter as user_config # type: ignore[assignment]
    except ImportError:
        raise ie # raise the original exception.. must be something else
    else:
        from ..core import warnings
        warnings.high('my.config.twitter is deprecated! Please rename it to my.config.twitter_archive in your config')
##


from dataclasses import dataclass
from datetime import datetime
from itertools import chain
import json  # hmm interesting enough, orjson didn't give much speedup here?
from pathlib import Path
from functools import cached_property
import html
from typing import (
    Iterator,
    List,
    Optional,
    Sequence,
)

from more_itertools import unique_everseen

from my.core import (
    datetime_aware,
    get_files,
    make_logger,
    stat,
    Json,
    Paths,
    Res,
    Stats,
)
from my.core import warnings
from my.core.cfg import make_config
from my.core.serialize import dumps as json_dumps

from .common import TweetId, permalink


@dataclass
class twitter_archive(user_config):
    export_path: Paths  # path[s]/glob to the twitter archive takeout


###

config = make_config(twitter_archive)


logger = make_logger(__name__)


def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


# TODO make sure it's not used anywhere else and simplify interface
@dataclass
class Tweet:
    raw: Json
    screen_name: str

    @property
    def id_str(self) -> TweetId:
        return self.raw['id_str']

    @property
    def created_at(self) -> datetime_aware:
        dts = self.raw['created_at']
        return datetime.strptime(dts, '%a %b %d %H:%M:%S %z %Y')

    @property
    def permalink(self) -> str:
        return permalink(screen_name=self.screen_name, id=self.id_str)

    @property
    def text(self) -> str:
        res: str = self.raw['full_text']

        ## replace shortened URLS
        repls = []  # from, to, what
        for ue in self.entities['urls']:
            [fr, to] = map(int, ue['indices'])
            repls.append((fr, to, ue['expanded_url']))
        # seems that media field isn't always set
        for me in self.entities.get('media', []):
            [fr, to] = map(int, me['indices'])
            repls.append((fr, to, me['display_url']))
            # todo not sure, maybe use media_url_https instead?
            # for now doing this for compatibility with twint
        repls = list(sorted(repls))
        parts = []
        idx = 0
        for fr, to, what in repls:
            parts.append(res[idx:fr])
            parts.append(what)
            idx = to
        parts.append(res[idx:])
        res = ''.join(parts)
        ##

        # replace stuff like &lt/&gt
        res = html.unescape(res)
        return res

    @property
    def urls(self) -> List[str]:
        ents = self.entities
        us = ents['urls']
        return [u['expanded_url'] for u in us]

    @property
    def entities(self) -> Json:
        # todo hmm what is 'extended_entities'
        return self.raw['entities']

    def __str__(self) -> str:
        return str(self.raw)

    def __repr__(self) -> str:
        return repr(self.raw)

    # TODO deprecate tid?
    @property
    def tid(self) -> TweetId:
        return self.id_str

    @property
    def dt(self) -> datetime_aware:
        return self.created_at


@dataclass
class Like:
    raw: Json
    screen_name: str

    @property
    def permalink(self) -> str:
        # doesn'tseem like link it export is more specific...
        return permalink(screen_name=self.screen_name, id=self.id_str)

    @property
    def id_str(self) -> TweetId:
        return self.raw['tweetId']

    @property
    def text(self) -> Optional[str]:
        # NOTE: likes basically don't have anything except text and url
        # ugh. I think none means that tweet was deleted?
        res: Optional[str] = self.raw.get('fullText')
        if res is None:
            return None
        res = html.unescape(res)
        return res

    # TODO deprecate?
    @property
    def tid(self) -> TweetId:
        return self.id_str


class ZipExport:
    def __init__(self, archive_path: Path) -> None:
        self.zpath = archive_path
        if (self.zpath / 'tweets.csv').exists():
            warnings.high("NOTE: CSV format (pre ~Aug 2018) isn't supported yet, this is likely not going to work.")
        self.old_format = False  # changed somewhere around 2020.03
        if not (self.zpath / 'Your archive.html').exists():
            self.old_format = True

    def raw(self, what: str, *, fname: Optional[str] = None) -> Iterator[Json]:
        logger.info(f'{self.zpath} : processing {what}')

        path = fname or what
        if not self.old_format:
            path = 'data/' + path
        path += '.js'

        ddd = (self.zpath / path).read_text()
        start = ddd.index('[')
        ddd = ddd[start:]
        for j in json.loads(ddd):
            if set(j.keys()) == {what}:
                # newer format
                yield j[what]
            else:
                # older format
                yield j

    @cached_property
    def screen_name(self) -> str:
        [acc] = self.raw(what='account')
        return acc['username']

    def tweets(self) -> Iterator[Tweet]:
        fname = 'tweets'  # since somewhere between mar and oct 2022
        if not (self.zpath / f'data/{fname}.js').exists():
            fname = 'tweet'  # old name
        # NOTE: for some reason, created_at doesn't seem to be in order
        # it mostly is, but there are a bunch of one-off random tweets where the time decreases (typically at the very end)
        for r in self.raw(what='tweet', fname=fname):
            yield Tweet(r, screen_name=self.screen_name)

    def likes(self) -> Iterator[Like]:
        # TODO ugh. would be nice to unify Tweet/Like interface
        # however, akeout only got tweetId, full text and url
        for r in self.raw(what='like'):
            yield Like(r, screen_name=self.screen_name)


# todo not sure about list and sorting? although can't hurt considering json is not iterative?
def tweets() -> Iterator[Res[Tweet]]:
    _all = chain.from_iterable(ZipExport(i).tweets() for i in inputs())
    res = unique_everseen(_all, key=json_dumps)
    yield from sorted(res, key=lambda t: t.dt)


def likes() -> Iterator[Res[Like]]:
    _all = chain.from_iterable(ZipExport(i).likes() for i in inputs())
    res = unique_everseen(_all, key=json_dumps)
    # ugh. likes don't have datetimes..
    yield from res


def stats() -> Stats:
    return {
        **stat(tweets),
        **stat(likes),
    }


## Deprecated stuff
Tid = TweetId



# this file only keeps the most common & critical types/utility functions
from .common import get_files, PathIsh, Paths
from .common import Json
from .common import warn_if_empty
from .common import stat, Stats
from .common import datetime_naive, datetime_aware
from .common import assert_never

from .cfg import make_config
from .error import Res, unwrap
from .logging import make_logger, LazyLogger
from .util import __NOT_HPI_MODULE__


# just for brevity in modules
# todo not sure about these.. maybe best to rely on regular imports.. perhaps compare?
from dataclasses import dataclass
from pathlib import Path


__all__ = [
    'get_files', 'PathIsh', 'Paths',
    'Json',
    'make_logger',
    'LazyLogger',  # legacy import
    'warn_if_empty',
    'stat', 'Stats',
    'datetime_aware', 'datetime_naive',
    'assert_never',

    'make_config',

    '__NOT_HPI_MODULE__',

    'Res', 'unwrap',

    'dataclass', 'Path',
]


## experimental for now
# you could put _init_hook.py next to your private my/config
# that way you can configure logging/warnings/env variables on every HPI import
try:
    import my._init_hook  # type: ignore[import-not-found]
except:
    pass
##



from my.core import __NOT_HPI_MODULE__

# NOTE: this tool was quite useful https://github.com/aj3423/aproto

from google.protobuf import descriptor_pool, descriptor_pb2, message_factory

TYPE_STRING  = descriptor_pb2.FieldDescriptorProto.TYPE_STRING
TYPE_BYTES   = descriptor_pb2.FieldDescriptorProto.TYPE_BYTES
TYPE_UINT64  = descriptor_pb2.FieldDescriptorProto.TYPE_UINT64
TYPE_MESSAGE = descriptor_pb2.FieldDescriptorProto.TYPE_MESSAGE

OPTIONAL = descriptor_pb2.FieldDescriptorProto.LABEL_OPTIONAL
REQUIRED = descriptor_pb2.FieldDescriptorProto.LABEL_REQUIRED


def get_place_protos():
    f1 = descriptor_pb2.DescriptorProto(name='xf1')
    # TODO 2 -> 5 is address? 2 -> 6 is a pair of coordinates
    f1.field.add(name='title', number=3, type=TYPE_STRING, label=REQUIRED)
    f1.field.add(name='note' , number=4, type=TYPE_STRING, label=OPTIONAL)
    # TODO what's the difference between required and optional? doesn't impact decoding?

    ts = descriptor_pb2.DescriptorProto(name='Timestamp')
    ts.field.add(name='seconds', number=1, type=TYPE_UINT64, label=REQUIRED)
    ts.field.add(name='nanos'  , number=2, type=TYPE_UINT64, label=REQUIRED)

    f1.field.add(name='created', number=10 ,type=TYPE_MESSAGE, label=REQUIRED, type_name=ts.name)
    f1.field.add(name='updated', number=11 ,type=TYPE_MESSAGE, label=REQUIRED, type_name=ts.name)

    f2 = descriptor_pb2.DescriptorProto(name='xf2')
    f2.field.add(name='addr1', number=2, type=TYPE_STRING, label=REQUIRED)
    f2.field.add(name='addr2', number=3, type=TYPE_STRING, label=REQUIRED)
    f2.field.add(name='f21'  , number=4, type=TYPE_BYTES , label=REQUIRED)
    f2.field.add(name='f22'  , number=5, type=TYPE_UINT64, label=REQUIRED)
    f2.field.add(name='f23'  , number=6, type=TYPE_STRING, label=REQUIRED)
    # NOTE: this also contains place ID

    f3 = descriptor_pb2.DescriptorProto(name='xf3')
    # NOTE: looks like it's the same as 'updated' from above??
    f3.field.add(name='f31', number=1, type=TYPE_UINT64, label=OPTIONAL)

    descriptor_proto = descriptor_pb2.DescriptorProto(name='PlaceParser')
    descriptor_proto.field.add(name='f1', number=1, type=TYPE_MESSAGE, label=REQUIRED, type_name=f1.name)
    descriptor_proto.field.add(name='f2', number=2, type=TYPE_MESSAGE, label=REQUIRED, type_name=f2.name)
    descriptor_proto.field.add(name='f3', number=3, type=TYPE_MESSAGE, label=OPTIONAL, type_name=f3.name)
    descriptor_proto.field.add(name='f4', number=4, type=TYPE_STRING , label=OPTIONAL)
    # NOTE: f4 is the list id

    return [descriptor_proto, ts, f1, f2, f3]


def get_labeled_protos():
    address = descriptor_pb2.DescriptorProto(name='address')
    # 1: address
    # 2: parts of address (multiple)
    # 3: full address
    address.field.add(name='full', number=3, type=TYPE_STRING, label=REQUIRED)

    main = descriptor_pb2.DescriptorProto(name='LabeledParser')
    # field 1 contains item type and item id
    main.field.add(name='title'  , number=3, type=TYPE_STRING , label=REQUIRED)
    main.field.add(name='address', number=5, type=TYPE_MESSAGE, label=OPTIONAL, type_name=address.name)

    return [main, address]


def get_list_protos():
    f1 = descriptor_pb2.DescriptorProto(name='xf1')
    f1.field.add(name='name', number=5, type=TYPE_STRING, label=REQUIRED)

    main = descriptor_pb2.DescriptorProto(name='ListParser')
    main.field.add(name='f1', number=1, type=TYPE_MESSAGE, label=REQUIRED, type_name=f1.name)
    main.field.add(name='f2', number=2, type=TYPE_STRING , label=REQUIRED)

    return [main, f1]


def make_parser(main, *extras):
    file_descriptor_proto = descriptor_pb2.FileDescriptorProto(name='dynamic.proto', package='dynamic_package')
    for proto in [main, *extras]:
        file_descriptor_proto.message_type.add().CopyFrom(proto)

    pool = descriptor_pool.DescriptorPool()
    file_descriptor = pool.Add(file_descriptor_proto)

    message_descriptor = pool.FindMessageTypeByName(f'{file_descriptor_proto.package}.{main.name}')
    factory = message_factory.MessageFactory(pool)
    dynamic_message_class = factory.GetPrototype(message_descriptor)

    return dynamic_message_class


place_parser_class   = make_parser(*get_place_protos())
labeled_parser_class = make_parser(*get_labeled_protos())
list_parser_class    = make_parser(*get_list_protos())


def parse_place(blob: bytes):
    m = place_parser_class()
    m.ParseFromString(blob)
    return m


def parse_labeled(blob: bytes):
    m = labeled_parser_class()
    m.ParseFromString(blob)
    return m


def parse_list(blob: bytes):
    msg = list_parser_class()
    msg.ParseFromString(blob)
    return msg



"""
Reddit data: saved items/comments/upvotes/etc.
"""
from __future__ import annotations

REQUIRES = [
    'git+https://github.com/karlicoss/rexport',
]

from dataclasses import dataclass
import inspect
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, Sequence

from my.core import (
    get_files,
    make_logger,
    warnings,
    stat,
    Paths,
    Stats,
)
from my.core.cfg import make_config, Attrs
from my.core.common import mcachew

from my.config import reddit as uconfig

logger = make_logger(__name__)


@dataclass
class reddit(uconfig):
    '''
    Uses [[https://github.com/karlicoss/rexport][rexport]] output.
    '''

    # path[s]/glob to the exported JSON data
    export_path: Paths


# hmm, also nice thing about this is that migration is possible to test without the rest of the config?
def migration(attrs: Attrs) -> Attrs:
    # new structure, take top-level config and extract 'rexport' class
    # previously, 'rexport' key could be location of the rexport repo on disk
    if 'rexport' in attrs and not isinstance(attrs['rexport'], (str, Path)):
        ex: uconfig.rexport = attrs['rexport']
        attrs['export_path'] = ex.export_path
    else:
        warnings.high(
            """DEPRECATED! Please modify your reddit config to look like:

class reddit:
    class rexport:
        export_path: Paths = '/path/to/rexport/data'
            """
        )
        export_dir = 'export_dir'
        if export_dir in attrs:  # legacy name
            attrs['export_path'] = attrs[export_dir]
            warnings.high(f'"{export_dir}" is deprecated! Please use "export_path" instead."')
    return attrs


config = make_config(reddit, migration=migration)

###
try:
    from rexport import dal
except ModuleNotFoundError as e:
    from my.core.hpi_compat import pre_pip_dal_handler

    dal = pre_pip_dal_handler('rexport', e, config, requires=REQUIRES)
# TODO ugh. this would import too early
# but on the other hand we do want to bring the objects into the scope for easier imports, etc. ugh!
# ok, fair enough I suppose. It makes sense to configure something before using it. can always figure it out later..
# maybe, the config could dynamically detect change and reimport itself? dunno.
###


def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


# TODO hmm so maybe these import here are not so great
# the issue is when the dal is updated (e.g. more types added)
# then user's state can be inconsistent if they update HPI, but don't update the dal
# maybe best to keep things begind the DAL after all

# fmt: off
Uid         = dal.Sid  # str
Save        = dal.Save
Comment     = dal.Comment
Submission  = dal.Submission
Upvote      = dal.Upvote
# fmt: on


def _dal() -> dal.DAL:
    sources = list(inputs())

    ## backwards compatibility (old rexport DAL didn't have cpu_pool argument)
    cpu_pool_arg = 'cpu_pool'
    pass_cpu_pool = cpu_pool_arg in inspect.signature(dal.DAL).parameters
    if pass_cpu_pool:
        from my.core._cpu_pool import get_cpu_pool

        kwargs = {cpu_pool_arg: get_cpu_pool()}
    else:
        kwargs = {}
    ##
    return dal.DAL(sources, **kwargs)


cache = mcachew(depends_on=inputs)


@cache
def saved() -> Iterator[Save]:
    return _dal().saved()


@cache
def comments() -> Iterator[Comment]:
    return _dal().comments()


@cache
def submissions() -> Iterator[Submission]:
    return _dal().submissions()


@cache
def upvoted() -> Iterator[Upvote]:
    return _dal().upvoted()


# uhh.. so with from __future__ import annotations, in principle we don't need updated export
# (with new entity types for function definitions below)
# however, cachew (as of 0.14.20231004) will crash during to get_type_hints call with these
# so we need to make cachew decorating defensive here
# will need to keep this for some time for backwards compatibility till cachew fix catches up
if not TYPE_CHECKING:
    # in runtime need to be defensive
    try:
        # here we just check that types are available, we don't actually want to import them
        # fmt: off
        dal.Subreddit
        dal.Profile
        dal.Multireddit
        # fmt: on
    except AttributeError as ae:
        warnings.high(f'{ae} : please update "rexport" installation')
        _cache = lambda f: f
        _USING_NEW_REXPORT = False
    else:
        _cache = cache
        _USING_NEW_REXPORT = True
else:
    _cache = cache


@_cache
def subreddits() -> Iterator[dal.Subreddit]:
    return _dal().subreddits()


@_cache
def multireddits() -> Iterator[dal.Multireddit]:
    return _dal().multireddits()


@_cache
def profile() -> dal.Profile:
    return _dal().profile()


def stats() -> Stats:
    return {
        # fmt: off
        **stat(saved      ),
        **stat(comments   ),
        **stat(submissions),
        **stat(upvoted    ),
        # fmt: on
    }



import sys
from typing import Any, Dict, Optional
import types


# The idea behind this one is to support accessing "overlaid/shadowed" modules from namespace packages
# See usage examples here:
# - https://github.com/karlicoss/hpi-personal-overlay/blob/master/src/my/util/hpi_heartbeat.py
# - https://github.com/karlicoss/hpi-personal-overlay/blob/master/src/my/twitter/all.py
# Suppose you want to use my.twitter.talon, which isn't in the default all.py
# You could just copy all.py to your personal overlay, but that would mean duplicating
# all the code and possible upstream changes.
# Alternatively, you could import the "original" my.twitter.all module from "overlay" my.twitter.all
# _ORIG = import_original_module(__name__, __file__)
# this would magically take care of package import path etc,
# and should import the "original" my.twitter.all as _ORIG
# After that you can call its methods, extend etc.
def import_original_module(
    module_name: str,
    file: str,
    *,
    star: bool = False,
    globals: Optional[Dict[str, Any]] = None,
) -> types.ModuleType:
    module_to_restore = sys.modules[module_name]

    # NOTE: we really wanna to hack the actual package of the module
    # rather than just top level my.
    # since that would be a bit less disruptive
    module_pkg = module_to_restore.__package__
    assert module_pkg is not None
    parent = sys.modules[module_pkg]

    my_path = parent.__path__._path  # type: ignore[attr-defined]
    my_path_orig = list(my_path)

    def fixup_path() -> None:
        for i, p in enumerate(my_path_orig):
            starts = file.startswith(p)
            if i == 0:
                # not sure about this.. but I guess it'll always be 0th element?
                assert starts, (my_path_orig, file)
            if starts:
                my_path.remove(p)
        # should remove exactly one item
        assert len(my_path) + 1 == len(my_path_orig), (my_path_orig, file)

    try:
        fixup_path()
        try:
            del sys.modules[module_name]
            # NOTE: we're using __import__ instead of importlib.import_module
            # since it's closer to the actual normal import (e.g. imports subpackages etc properly )
            # fromlist=[None] forces it to return rightmost child
            # (otherwise would just return 'my' package)
            res = __import__(module_name, fromlist=[None])  # type: ignore[list-item]
            if star:
                assert globals is not None
                globals.update({k: v for k, v in vars(res).items() if not k.startswith('_')})
            return res
        finally:
            sys.modules[module_name] = module_to_restore
    finally:
        my_path[:] = my_path_orig



# todo: uses my private export script?
from datetime import datetime
import json
from typing import NamedTuple, Iterable, Sequence, Optional


from my.config import vk as config  # type: ignore[attr-defined]


class Favorite(NamedTuple):
    dt: datetime
    title: str
    url: Optional[str]
    text: str


from ..core import Json
from ..core.error import Res


skip = (
    'graffiti',
    'poll',

    # TODO could be useful..
    'note',
    'doc',
    'audio',
    'photo',
    'album',
    'video',
    'page',
)

def parse_fav(j: Json) -> Favorite:
    # TODO copy_history??
    url = None
    title = '' # TODO ???
    atts = j.get('attachments', [])
    for a in atts:
        if any(k in a for k in skip):
            continue
        link = a['link']
        title = link['title']
        url = link['url']
        #  TODOlink['description'] ?

    # TODO would be nice to include user
    return Favorite(
        dt=datetime.utcfromtimestamp(j['date']),
        title=title,
        url=url,
        text=j['text'],
    )


def _iter_favs() -> Iterable[Res]:
    jj = json.loads(config.favs_file.read_text())
    for j in jj:
        try:
            yield parse_fav(j)
        except Exception as e:
            ex = RuntimeError(f"Error while processing\n{j}")
            ex.__cause__ = e
            yield ex


def favorites() -> Sequence[Res]:
    it = _iter_favs()
    # trick to sort errors along with the actual objects
    # TODO wonder if there is a shorter way?
    # TODO add to the error handling post?
    favs = list(it)
    prev = datetime.min
    keys = []
    for i, f in enumerate(favs):
        if not isinstance(f, Exception):
            prev = f.dt
        keys.append((prev, i)) # include index to resolve ties
    sorted_items = [p[1] for p in sorted(zip(keys, favs))]
    #
    return sorted_items


def stats():
    from ..core import stat
    return stat(favorites)



"""
Parse [[https://github.com/mendhak/gpslogger][gpslogger]] .gpx (xml) files
"""

REQUIRES = ["gpxpy"]

from my.config import location
from my.core import Paths, dataclass


@dataclass
class config(location.gpslogger):
    # path[s]/glob to the synced gpx (XML) files
    export_path: Paths

    # default accuracy for gpslogger
    accuracy: float = 50.0


from itertools import chain
from datetime import datetime, timezone
from pathlib import Path
from typing import Iterator, Sequence, List

import gpxpy
from gpxpy.gpx import GPXXMLSyntaxException
from more_itertools import unique_everseen

from my.core import Stats, LazyLogger
from my.core.common import get_files, mcachew
from .common import Location


logger = LazyLogger(__name__, level="warning")

def _input_sort_key(path: Path) -> str:
    if "_" in path.name:
        return path.name.split("_", maxsplit=1)[1]
    return path.name


def inputs() -> Sequence[Path]:
    # gpslogger files can optionally be prefixed by a device id,
    # like b5760c66102a5269_20211214142156.gpx
    return sorted(get_files(config.export_path, glob="*.gpx", sort=False), key=_input_sort_key)


def _cachew_depends_on() -> List[float]:
    return [p.stat().st_mtime for p in inputs()]


# TODO: could use a better cachew key/this has to recompute every file whenever the newest one changes
@mcachew(depends_on=_cachew_depends_on, logger=logger)
def locations() -> Iterator[Location]:
    yield from unique_everseen(
        chain(*map(_extract_locations, inputs())), key=lambda loc: loc.dt
    )


def _extract_locations(path: Path) -> Iterator[Location]:
    with path.open("r") as gf:
        try:
            gpx_obj = gpxpy.parse(gf)
        except GPXXMLSyntaxException as e:
            logger.warning("failed to parse XML %s: %s", path, e)
            return
        for track in gpx_obj.tracks:
            for segment in track.segments:
                for point in segment.points:
                    if point.time is None:
                        continue
                    # hmm - for gpslogger, seems that timezone is always SimpleTZ('Z'), which
                    # specifies UTC -- see https://github.com/tkrajina/gpxpy/blob/cb243b22841bd2ce9e603fe3a96672fc75edecf2/gpxpy/gpxfield.py#L38
                    yield Location(
                        lat=point.latitude,
                        lon=point.longitude,
                        accuracy=config.accuracy,
                        elevation=point.elevation,
                        dt=datetime.replace(point.time, tzinfo=timezone.utc),
                        datasource="gpslogger",
                    )


def stats() -> Stats:
    from my.core import stat

    return {**stat(locations)}



from my.tests.common import skip_if_not_karlicoss as pytestmark

from my import orgmode
from my.core.orgmode import collect


def test() -> None:
    # meh
    results = list(orgmode.query().collect_all(lambda n: [n] if 'python' in n.tags else []))
    assert len(results) > 5



"""
Decorator to gracefully handle importing a data source, or warning
and yielding nothing (or a default) when its not available
"""

from functools import wraps
from typing import Any, Iterator, TypeVar, Callable, Optional, Iterable
import warnings

from .warnings import medium

# The factory function may produce something that has data
# similar to the shared model, but not exactly, so not
# making this a TypeVar, is just to make reading the
# type signature below a bit easier...
T = Any

# https://mypy.readthedocs.io/en/latest/generics.html?highlight=decorators#decorator-factories
FactoryF = TypeVar("FactoryF", bound=Callable[..., Iterator[T]])

_DEFAULT_ITR = ()


# tried to use decorator module but it really doesn't work well
# with types and kw-arguments... :/
def import_source(
    *,
    default: Iterable[T] = _DEFAULT_ITR,
    module_name: Optional[str] = None,
    help_url: Optional[str] = None,
) -> Callable[..., Callable[..., Iterator[T]]]:
    """
    doesn't really play well with types, but is used to catch
    ModuleNotFoundError's for when modules aren't installed in
    all.py files, so the types don't particularly matter

    this is meant to be used to wrap some function which imports
    and then yields an iterator of objects

    If the user doesn't have that module installed, it returns
    nothing and warns instead
    """

    def decorator(factory_func: FactoryF) -> Callable[..., Iterator[T]]:
        @wraps(factory_func)
        def wrapper(*args, **kwargs) -> Iterator[T]:
            try:
                res = factory_func(*args, **kwargs)
                yield from res
            except (ImportError, AttributeError) as err:
                from . import core_config as CC
                from .error import warn_my_config_import_error
                suppressed_in_conf = False
                if module_name is not None and CC.config._is_module_active(module_name) is False:
                    suppressed_in_conf = True
                if not suppressed_in_conf:
                    if module_name is None:
                        medium(f"Module {factory_func.__qualname__} could not be imported, or isn't configured properly")
                    else:
                        medium(f"Module {module_name} ({factory_func.__qualname__}) could not be imported, or isn't configured properly")
                        warnings.warn(f"""If you don't want to use this module, to hide this message, add '{module_name}' to your core config disabled_modules in your config, like:

class core:
    disabled_modules = [{repr(module_name)}]
""")
                    # try to check if this is a config error or based on dependencies not being installed
                    if isinstance(err, (ImportError, AttributeError)):
                        matched_config_err = warn_my_config_import_error(err, help_url=help_url)
                        # if we determined this wasn't a config error, and it was an attribute error
                        # it could be *any* attribute error -- we should raise this since its otherwise a fatal error
                        # from some code in the module failing
                        if not matched_config_err and isinstance(err, AttributeError):
                            raise err
                yield from default
        return wrapper
    return decorator



from my.core import __NOT_HPI_MODULE__
from my.core import warnings

warnings.high('my.kython.kompress is deprecated, please use "kompress" library directly. See https://github.com/karlicoss/kompress')

from my.core.kompress import *



from my.tests.common import skip_if_not_karlicoss as pytestmark
# TODO maybe instead detect if it has any data at all
# if none, then skip the test, say that user doesn't have any data?

# TODO implement via stat?
def test() -> None:
    from my.smscalls import calls, messages, mms
    assert len(list(calls())) > 10
    assert len(list(messages())) > 10
    assert len(list(mms())) > 10



"""
Zulip data from [[https://memex.zulipchat.com/help/export-your-organization][Organization export]]
"""
from dataclasses import dataclass
from datetime import datetime, timezone
from itertools import count
import json
from pathlib import Path
from typing import Sequence, Iterator, Dict, Union

from my.core import (
    assert_never,
    datetime_aware,
    get_files,
    stat,
    Json,
    Paths,
    Res,
    Stats,
)
from my.core.error import notnone
import my.config


@dataclass
class organization(my.config.zulip.organization):
    # paths[s]/glob to the exported JSON data
    export_path: Paths


def inputs() -> Sequence[Path]:
    # TODO: seems like export ids are kinda random..
    # not sure what's the best way to figure out the last without renaming?
    # could use mtime perhaps?
    return get_files(organization.export_path, sort=False)


@dataclass(frozen=True)
class Server:
    id: int
    string_id: str
    name: str


@dataclass(frozen=True)
class Sender:
    id: int
    # todo make optional?
    full_name: str
    email: str


# from the data, seems that subjects are completely implicit and determined by name?
# streams have ids (can extract from realm/zerver_stream), but unclear how to correlate messages/topics to streams?
@dataclass(frozen=True)
class _Message:
    # todo hmm not sure what would be a good field order..
    id: int
    sent: datetime_aware  # double checked and they are in utc
    subject: str
    sender_id: int
    server_id: int
    content: str  # TODO hmm, it keeps markdown, not sure how/whether it's worth to prettify at all?
    # TODO recipient??
    # todo keep raw item instead? not sure


@dataclass(frozen=True)
class Message:
    id: int
    sent: datetime_aware
    subject: str
    sender: Sender
    server: Server
    content: str

    @property
    def permalink(self) -> str:
        # seems that these link to the same message
        # https://memex.zulipchat.com/#narrow/stream/284580-python/topic/py-spy.20profiler/near/234798881
        # https://memex.zulipchat.com/#narrow/stream/284580/near/234798881
        # https://memex.zulipchat.com/#narrow/near/234798881
        # however not sure how to correlate stream id and message/topic for now, so preferring the latter version
        return f'https://{self.server.string_id}.zulipchat.com/#narrow/near/{self.id}'


# todo cache it
def _entities() -> Iterator[Res[Union[Server, Sender, _Message]]]:
    last = max(inputs())

    # todo would be nice to switch it to unpacked dirs as well, similar to ZipPath
    # I guess makes sense to have a special implementation for .tar.gz considering how common are they
    import tarfile

    tfile = tarfile.open(last)

    subdir = tfile.getnames()[0]  # there is a directory inside tar file, first name should be that

    with notnone(tfile.extractfile(f'{subdir}/realm.json')) as fo:
        rj = json.load(fo)

    [sj] = rj['zerver_realm']
    server = Server(
        id=sj['id'],
        string_id=sj['string_id'],
        name=sj['name'],
    )
    yield server

    for j in rj['zerver_userprofile']:
        yield Sender(
            id=j['id'],
            full_name=j['full_name'],
            email=j['email'],
        )

    for j in rj['zerver_userprofile_crossrealm']:  # e.g. zulip bot
        yield Sender(
            id=j['id'],
            full_name=j['email'],  # doesn't seem to have anything
            email=j['email'],
        )

    def _parse_message(j: Json) -> _Message:
        ds = j['date_sent']
        # fmt: off
        return _Message(
            id        = j['id'],
            sent      = datetime.fromtimestamp(ds, tz=timezone.utc),
            subject   = j['subject'],
            sender_id = j['sender'],
            server_id = server.id,
            content   = j['content'],
        )
        # fmt: on

    for idx in count(start=1, step=1):
        fname = f'messages-{idx:06}.json'
        fpath = f'{subdir}/{fname}'
        if fpath not in tfile.getnames():
            # tarfile doesn't have .exists?
            break
        with notnone(tfile.extractfile(fpath)) as fo:
            mj = json.load(fo)
        # TODO handle  zerver_usermessage
        for j in mj['zerver_message']:
            try:
                yield _parse_message(j)
            except Exception as e:
                yield e


def messages() -> Iterator[Res[Message]]:
    id2sender: Dict[int, Sender] = {}
    id2server: Dict[int, Server] = {}
    for x in _entities():
        if isinstance(x, Exception):
            yield x
            continue
        if isinstance(x, Server):
            id2server[x.id] = x
            continue
        if isinstance(x, Sender):
            id2sender[x.id] = x
            continue
        if isinstance(x, _Message):
            # TODO a bit copypasty... wonder if possible to mixin or something instead
            yield Message(
                id=x.id,
                sent=x.sent,
                subject=x.subject,
                sender=id2sender[x.sender_id],
                server=id2server[x.server_id],
                content=x.content,
            )
            continue
        assert_never(x)


def stats() -> Stats:
    return {**stat(messages)}



from setuptools import setup, find_namespace_packages # type: ignore


def main() -> None:
    pkgs = find_namespace_packages('src')
    pkg = min(pkgs)
    setup(
        name='hpi-overlay3',
        zip_safe=False,
        packages=pkgs,
        package_dir={'': 'src'},
        package_data={pkg: ['py.typed']},
    )


if __name__ == '__main__':
    main()



"""
To test my.location.fallback_location.all
"""

from typing import Iterator
from datetime import datetime, timezone, timedelta

from more_itertools import ilen

from my.ip.common import IP

def data() -> Iterator[IP]:
    # random IP addresses
    yield IP(addr="67.98.113.0", dt=datetime(2020, 1, 1, 12, 0, 0, tzinfo=timezone.utc))
    yield IP(addr="67.98.112.0", dt=datetime(2020, 1, 15, 12, 0, 0, tzinfo=timezone.utc))
    yield IP(addr="59.40.113.87", dt=datetime(2020, 2, 1, 12, 0, 0, tzinfo=timezone.utc))
    yield IP(addr="59.40.139.87", dt=datetime(2020, 2, 1, 16, 0, 0, tzinfo=timezone.utc))
    yield IP(addr="161.235.192.228", dt=datetime(2020, 3, 1, 12, 0, 0, tzinfo=timezone.utc))

# redefine the my.ip.all function using data for testing
import my.ip.all as ip_module
ip_module.ips = data

from my.location.fallback import via_ip

# these are all tests for the bisect algorithm defined in via_ip.py
# to make sure we can correctly find IPs that are within the 'for_duration' of a given datetime

def test_ip_fallback() -> None:
    # make sure that the data override works
    assert ilen(ip_module.ips()) == ilen(data())
    assert ilen(ip_module.ips()) == ilen(via_ip.fallback_locations())
    assert ilen(via_ip.fallback_locations()) == 5
    assert ilen(via_ip._sorted_fallback_locations()) == 5

    # confirm duration from via_ip since that is used for bisect
    assert via_ip.config.for_duration == timedelta(hours=24)

    # basic tests

    # try estimating slightly before the first IP
    est = list(via_ip.estimate_location(datetime(2020, 1, 1, 11, 59, 59, tzinfo=timezone.utc)))
    assert len(est) == 0

    # during the duration for the first IP
    est = list(via_ip.estimate_location(datetime(2020, 1, 1, 12, 30, 0, tzinfo=timezone.utc)))
    assert len(est) == 1

    # right after the 'for_duration' for an IP
    est = list(via_ip.estimate_location(datetime(2020, 1, 1, 12, 0, 0, tzinfo=timezone.utc) + via_ip.config.for_duration + timedelta(seconds=1)))
    assert len(est) == 0

    # on 2/1/2020, threes one IP if before 16:30
    est = list(via_ip.estimate_location(datetime(2020, 2, 1, 12, 30, 0, tzinfo=timezone.utc)))
    assert len(est) == 1

    # and two if after 16:30
    est = list(via_ip.estimate_location(datetime(2020, 2, 1, 17, 00, 0, tzinfo=timezone.utc)))
    assert len(est) == 2

    # the 12:30 IP should 'expire' before the 16:30 IP, use 3:30PM on the next day
    est = list(via_ip.estimate_location(datetime(2020, 2, 2, 15, 30, 0, tzinfo=timezone.utc)))
    assert len(est) == 1

    use_dt = datetime(2020, 3, 1, 12, 15, 0, tzinfo=timezone.utc)

    # test last IP
    est = list(via_ip.estimate_location(use_dt))
    assert len(est) == 1

    # datetime should be the IPs, not the passed IP (if via_home, it uses the passed dt)
    assert est[0].dt != use_dt

    # test interop with other fallback estimators/all.py
    #
    # redefine fallback_estimators to prevent possible namespace packages the user
    # may have installed from having side effects testing this
    from my.location.fallback import all
    from my.location.fallback import via_home
    def _fe() -> Iterator[all.LocationEstimator]:
        yield via_ip.estimate_location
        yield via_home.estimate_location

    all.fallback_estimators = _fe
    assert ilen(all.fallback_estimators()) == 2

    # test that all.estimate_location has access to both IPs
    #
    # just passing via_ip should give one IP
    from my.location.fallback.common import _iter_estimate_from
    raw_est = list(_iter_estimate_from(use_dt, (via_ip.estimate_location,)))
    assert len(raw_est) == 1
    assert raw_est[0].datasource == "via_ip"
    assert raw_est[0].accuracy == 15_000

    # passing home should give one
    home_est = list(_iter_estimate_from(use_dt, (via_home.estimate_location,)))
    assert len(home_est) == 1
    assert home_est[0].accuracy == 30_000

    # make sure ip accuracy is more accurate
    assert raw_est[0].accuracy < home_est[0].accuracy

    # passing both should give two
    raw_est = list(_iter_estimate_from(use_dt, (via_ip.estimate_location, via_home.estimate_location)))
    assert len(raw_est) == 2

    # shouldn't raise value error
    all_est = all.estimate_location(use_dt)
    # should have used the IP from via_ip since it was more accurate
    assert all_est.datasource == "via_ip"

    # test that a home defined in shared_config.py is used if no IP is found
    loc = all.estimate_location(datetime(2021, 1, 1, 12, 30, 0, tzinfo=timezone.utc))
    assert loc.datasource == "via_home"

    # test a different home using location.fallback.all
    bulgaria = all.estimate_location(datetime(2006, 1, 1, 12, 30, 0, tzinfo=timezone.utc))
    assert bulgaria.datasource == "via_home"
    assert (bulgaria.lat, bulgaria.lon) == (42.697842, 23.325973)
    assert (loc.lat, loc.lon) != (bulgaria.lat, bulgaria.lon)


# re-use prepare fixture for overriding config from shared_config.py
from .tz import prepare



'''
Module for locating and accessing [[https://takeout.google.com][Google Takeout]] data
'''

from dataclasses import dataclass
from ...core.common import Paths, get_files
from ...core.util import __NOT_HPI_MODULE__

from my.config import google as user_config

from more_itertools import last

@dataclass
class google(user_config):
    takeout_path: Paths # path/paths/glob for the takeout zips
###

# TODO rename 'google' to 'takeout'? not sure

from ...core.cfg import make_config
config = make_config(google)

from pathlib import Path
from typing import Optional, Iterable


def get_takeouts(*, path: Optional[str]=None) -> Iterable[Path]:
    """
    Sometimes google splits takeout into multiple archives, so we need to detect the ones that contain the path we need
    """
    # TODO FIXME zip is not great..
    # allow a lambda expression? that way the user could restrict it
    for takeout in get_files(config.takeout_path, glob='*.zip'):
        if path is None or (takeout / path).exists():
            yield takeout


def get_last_takeout(*, path: Optional[str]=None) -> Optional[Path]:
    return last(get_takeouts(path=path), default=None)


# TODO might be a good idea to merge across multiple takeouts...
# perhaps even a special takeout module that deals with all of this automatically?
# e.g. accumulate, filter and maybe report useless takeouts?




import sys
from datetime import datetime, timedelta
from pathlib import Path

import pytest
import pytz

from my.core.error import notnone

import my.time.tz.main as TZ
import my.time.tz.via_location as LTZ


def test_iter_tzs() -> None:
    ll = list(LTZ._iter_tzs())
    assert len(ll) > 3


def test_past() -> None:
    # should fallback to the home location provider
    dt = D('20000101 12:34:45')
    dt = TZ.localize(dt)
    tz = dt.tzinfo
    assert tz is not None
    assert getattr(tz, 'zone') == 'America/New_York'


def test_future() -> None:
    fut = datetime.now() + timedelta(days=100)
    # shouldn't crash at least
    assert TZ.localize(fut) is not None


def test_tz() -> None:
    # todo hmm, the way it's implemented at the moment, never returns None?

    # not present in the test data
    tz = LTZ._get_tz(D('20200101 10:00:00'))
    assert notnone(tz).zone == 'Europe/Sofia'

    tz = LTZ._get_tz(D('20170801 11:00:00'))
    assert notnone(tz).zone == 'Europe/Vienna'

    tz = LTZ._get_tz(D('20170730 10:00:00'))
    assert notnone(tz).zone == 'Europe/Rome'

    tz = LTZ._get_tz(D('20201001 14:15:16'))
    assert tz is not None

    on_windows = sys.platform == 'win32'
    if not on_windows:
        tz = LTZ._get_tz(datetime.min)
        assert tz is not None
    else:
        # seems this fails because windows doesn't support same date ranges
        # https://stackoverflow.com/a/41400321/
        with pytest.raises(OSError):
            LTZ._get_tz(datetime.min)


def test_policies() -> None:
    getzone = lambda dt: getattr(dt.tzinfo, 'zone')

    naive = D('20170730 10:00:00')
    # actual timezone at the time
    assert getzone(TZ.localize(naive)) == 'Europe/Rome'

    z = pytz.timezone('America/New_York')
    aware = z.localize(naive)

    assert getzone(TZ.localize(aware)) == 'America/New_York'

    assert getzone(TZ.localize(aware, policy='convert')) == 'Europe/Rome'


    with pytest.raises(RuntimeError):
        assert TZ.localize(aware, policy='throw')


def D(dstr: str) -> datetime:
    return datetime.strptime(dstr, '%Y%m%d %H:%M:%S')



@pytest.fixture(autouse=True)
def prepare(tmp_path: Path):
    from .shared_config import temp_config
    conf = temp_config(tmp_path)

    import my.core.cfg as C
    with C.tmp_config() as config:
        config.google   = conf.google
        config.time     = conf.time
        config.location = conf.location
        yield



"""
Gives you access to older comments possibly not accessible with rexport
using pushshift
See https://github.com/seanbreckenridge/pushshift_comment_export
"""

REQUIRES = [
    "git+https://github.com/seanbreckenridge/pushshift_comment_export",
]

from my.core.common import Paths, Stats
from dataclasses import dataclass
from my.core.cfg import make_config

# note: keeping pushshift import before config import, so it's handled gracefully by import_source
from pushshift_comment_export.dal import read_file, PComment

from my.config import reddit as uconfig

@dataclass
class pushshift_config(uconfig.pushshift):
    '''
    Uses [[https://github.com/seanbreckenridge/pushshift_comment_export][pushshift]] to get access to old comments
    '''

    # path[s]/glob to the exported JSON data
    export_path: Paths

config = make_config(pushshift_config)

from my.core import get_files
from typing import Sequence, Iterator
from pathlib import Path



def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


def comments() -> Iterator[PComment]:
    for f in inputs():
        yield from read_file(f)

def stats() -> Stats:
    from my.core import stat
    return {
        **stat(comments)
    }




from my.tests.common import skip_if_not_karlicoss as pytestmark

def test_checkins() -> None:
    from my.foursquare import get_checkins
    # todo reuse stats?
    checkins = get_checkins()
    assert len(checkins) > 100
    assert any('Victoria Park' in c.summary for c in checkins)
    # TODO cid_map??



from .common import assert_subpackage; assert_subpackage(__name__)


from contextlib import contextmanager
from pathlib import Path
import shutil
import sqlite3
from tempfile import TemporaryDirectory
from typing import Tuple, Any, Iterator, Callable, Optional, Union, Literal


from .common import PathIsh, assert_never


def sqlite_connect_immutable(db: PathIsh) -> sqlite3.Connection:
    return sqlite3.connect(f'file:{db}?immutable=1', uri=True)


def test_sqlite_connect_immutable(tmp_path: Path) -> None:
    db = str(tmp_path / 'db.sqlite')
    with sqlite3.connect(db) as conn:
        conn.execute('CREATE TABLE testtable (col)')

    import pytest
    with pytest.raises(sqlite3.OperationalError, match='readonly database'):
        with sqlite_connect_immutable(db) as conn:
            conn.execute('DROP TABLE testtable')

    # succeeds without immutable
    with sqlite3.connect(db) as conn:
        conn.execute('DROP TABLE testtable')


SqliteRowFactory = Callable[[sqlite3.Cursor, sqlite3.Row], Any]

def dict_factory(cursor, row):
    fields = [column[0] for column in cursor.description]
    return {key: value for key, value in zip(fields, row)}


Factory = Union[SqliteRowFactory, Literal['row', 'dict']]

@contextmanager
def sqlite_connection(db: PathIsh, *, immutable: bool=False, row_factory: Optional[Factory]=None) -> Iterator[sqlite3.Connection]:
    dbp = f'file:{db}'
    # https://www.sqlite.org/draft/uri.html#uriimmutable
    if immutable:
        # assert results in nicer error than sqlite3.OperationalError
        assert Path(db).exists(), db
        dbp = f'{dbp}?immutable=1'
    row_factory_: Any = None
    if row_factory is not None:
        if callable(row_factory):
            row_factory_ = row_factory
        elif row_factory == 'row':
            row_factory_ = sqlite3.Row
        elif row_factory == 'dict':
            row_factory_ = dict_factory
        else:
            assert_never()

    conn = sqlite3.connect(dbp, uri=True)
    try:
        conn.row_factory = row_factory_
        with conn:
            yield conn
    finally:
        # Connection context manager isn't actually closing the connection, only keeps transaction
        conn.close()


# TODO come up with a better name?
# NOTE: this is tested by tests/sqlite.py::test_sqlite_read_with_wal
def sqlite_copy_and_open(db: PathIsh) -> sqlite3.Connection:
    """
    'Snapshots' database and opens by making a deep copy of it including journal/WAL files
    """
    dp = Path(db)
    # TODO make atomic/check mtimes or something
    dest = sqlite3.connect(':memory:')
    with TemporaryDirectory() as td:
        tdir = Path(td)
        # shm should be recreated from scratch -- safer not to copy perhaps
        tocopy = [dp] + [p for p in dp.parent.glob(dp.name + '-*') if not p.name.endswith('-shm')]
        for p in tocopy:
            shutil.copy(p, tdir / p.name)
        with sqlite3.connect(str(tdir / dp.name)) as conn:
            conn.backup(target=dest)
        conn.close()
    return dest


# NOTE hmm, so this kinda works
# V = TypeVar('V', bound=Tuple[Any, ...])
# def select(cols: V, rest: str, *, db: sqlite3.Connection) -> Iterator[V]:
# but sadly when we pass columns (Tuple[str, ...]), it seems to bind this type to V?
# and then the return type ends up as Iterator[Tuple[str, ...]], which isn't desirable :(
# a bit annoying to have this copy-pasting, but hopefully not a big issue

from typing import overload
@overload
def select(cols: Tuple[str                                   ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[Tuple[Any                                   ]]: ...
@overload
def select(cols: Tuple[str, str                              ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[Tuple[Any, Any                              ]]: ...
@overload
def select(cols: Tuple[str, str, str                         ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[Tuple[Any, Any, Any                         ]]: ...
@overload
def select(cols: Tuple[str, str, str, str                    ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[Tuple[Any, Any, Any, Any                    ]]: ...
@overload
def select(cols: Tuple[str, str, str, str, str               ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[Tuple[Any, Any, Any, Any, Any               ]]: ...
@overload
def select(cols: Tuple[str, str, str, str, str, str          ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[Tuple[Any, Any, Any, Any, Any, Any          ]]: ...
@overload
def select(cols: Tuple[str, str, str, str, str, str, str     ], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[Tuple[Any, Any, Any, Any, Any, Any, Any     ]]: ...
@overload
def select(cols: Tuple[str, str, str, str, str, str, str, str], rest: str, *, db: sqlite3.Connection) -> \
        Iterator[Tuple[Any, Any, Any, Any, Any, Any, Any, Any]]: ...

def select(cols, rest, *, db):
    # db arg is last cause that results in nicer code formatting..
    return db.execute('SELECT ' + ','.join(cols) + ' ' + rest)



from my.tests.common import skip_if_not_karlicoss as pytestmark


def test_pages() -> None:
    # TODO ugh. need lazy import to simplify testing?
    from my.instapaper import pages
    assert len(list(pages())) > 3



from pathlib import Path

import pytest


def test() -> None:
    from my.location.google import locations
    locs = list(locations())
    assert len(locs) == 3810

    last = locs[-1]
    assert last.dt.strftime('%Y%m%d %H:%M:%S') == '20170802 13:01:56' # should be utc
    # todo approx
    assert last.lat == 46.5515350
    assert last.lon == 16.4742742
    # todo check altitude


@pytest.fixture(autouse=True)
def prepare(tmp_path: Path):
    from .shared_config import temp_config
    user_config = temp_config(tmp_path)

    import my.core.cfg as C
    with C.tmp_config() as config:
        config.google = user_config.google
        yield




from typing import Iterator

from my.core import Stats
from my.core.source import import_source
from browserexport.merge import merge_visits, Visit


src_export = import_source(module_name="my.browser.export")
src_active = import_source(module_name="my.browser.active_browser")


@src_export
def _visits_export() -> Iterator[Visit]:
    from . import export
    return export.history()


@src_active
def _visits_active() -> Iterator[Visit]:
    from . import active_browser
    return active_browser.history()


# NOTE: you can comment out the sources you don't need
def history() -> Iterator[Visit]:
    yield from merge_visits([
        _visits_active(),
        _visits_export(),
    ])


def stats() -> Stats:
    from my.core import stat

    return {**stat(history)}



'''
TODO doesn't really belong to 'core' morally, but can think of moving out later
'''
from .common import assert_subpackage; assert_subpackage(__name__)

from typing import Iterable, Any, Optional, Dict

from .common import LazyLogger, asdict, Json


logger = LazyLogger(__name__)


class config:
    db = 'db'


RESET_DEFAULT = False


def fill(it: Iterable[Any], *, measurement: str, reset: bool=RESET_DEFAULT, dt_col: str='dt') -> None:
    # todo infer dt column automatically, reuse in stat?
    # it doesn't like dots, ends up some syntax error?
    measurement = measurement.replace('.', '_')
    # todo autoinfer measurement?

    db = config.db

    from influxdb import InfluxDBClient # type: ignore
    client = InfluxDBClient()
    # todo maybe create if not exists?
    # client.create_database(db)

    # todo should be it be env variable?
    if reset:
        logger.warning('deleting measurements: %s:%s', db, measurement)
        client.delete_series(database=db, measurement=measurement)

    # TODO need to take schema here...
    cache: Dict[str, bool] = {}

    def good(f, v) -> bool:
        c = cache.get(f)
        if c is not None:
            return c
        t = type(v)
        r = t in {str, int}
        cache[f] = r
        if not r:
            logger.warning('%s: filtering out %s=%s because of type %s', measurement, f, v, t)
        return r

    def filter_dict(d: Json) -> Json:
        return {f: v for f, v in d.items() if good(f, v)}

    def dit() -> Iterable[Json]:
        for i in it:
            d = asdict(i)
            tags: Optional[Json] = None
            tags_ = d.get('tags') # meh... handle in a more robust manner
            if tags_ is not None and isinstance(tags_, dict): # FIXME meh.
                del d['tags']
                tags = tags_

            # TODO what to do with exceptions??
            # todo handle errors.. not sure how? maybe add tag for 'error' and fill with empty data?
            dt = d[dt_col].isoformat()
            del d[dt_col]

            fields = filter_dict(d)

            yield dict(
                measurement=measurement,
                # TODO maybe good idea to tag with database file/name? to inspect inconsistencies etc..
                # hmm, so tags are autoindexed and might be faster?
                # not sure what's the big difference though
                # "fields are data and tags are metadata"
                tags=tags,
                time=dt,
                fields=fields,
            )

    from more_itertools import chunked
    # "The optimal batch size is 5000 lines of line protocol."
    # some chunking is def necessary, otherwise it fails
    inserted = 0
    for chi in chunked(dit(), n=5000):
        chl = list(chi)
        inserted += len(chl)
        logger.debug('writing next chunk %s', chl[-1])
        client.write_points(chl, database=db)

    logger.info('inserted %d points', inserted)
    # todo "Specify timestamp precision when writing to InfluxDB."?


def magic_fill(it, *, name: Optional[str]=None, reset: bool=RESET_DEFAULT) -> None:
    if name is None:
        assert callable(it) # generators have no name/module
        name = f'{it.__module__}:{it.__name__}'
    assert name is not None

    if callable(it):
        it = it()

    from itertools import tee
    from more_itertools import first, one
    it, x = tee(it)
    f = first(x, default=None)
    if f is None:
        logger.warning('%s has no data', name)
        return

    # TODO can we reuse pandas code or something?
    #
    from .pandas import _as_columns
    schema = _as_columns(type(f))

    from datetime import datetime
    dtex = RuntimeError(f'expected single datetime field. schema: {schema}')
    dtf = one((f for f, t in schema.items() if t == datetime), too_short=dtex, too_long=dtex)

    fill(it, measurement=name, reset=reset, dt_col=dtf)


import click

@click.group()
def main() -> None:
    pass


@main.command(name='populate', short_help='populate influxdb')
@click.option('--reset', is_flag=True, help='Reset Influx measurements before inserting', show_default=True)
@click.argument('FUNCTION_NAME', type=str, required=True)
def populate(function_name: str, reset: bool) -> None:
    from .__main__ import _locate_functions_or_prompt
    [provider] = list(_locate_functions_or_prompt([function_name]))
    # todo could have a non-interactive version which populates from all data sources for the provider?
    magic_fill(provider, reset=reset)


# todo later just add to hpi main?
# not sure if want to couple
if __name__ == '__main__':
    main()



"""
Extracts data from the official Google Maps app for Android (uses gmm_sync.db for now)
"""
from __future__ import annotations

REQUIRES = [
    "protobuf",  # for parsing blobs from the database
]

from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Iterator, Optional, Sequence
from urllib.parse import quote

from my.core import datetime_aware, get_files, LazyLogger, Paths, Res
from my.core.common import unique_everseen
from my.core.sqlite import sqlite_connection

import my.config

from ._android_protobuf import parse_labeled, parse_list, parse_place


logger = LazyLogger(__name__)


@dataclass
class config(my.config.google.maps.android):
    # paths[s]/glob to the exported sqlite databases
    export_path: Paths


def inputs() -> Sequence[Path]:
    # TODO note sure if need to use all dbs? possibly the last one contains everything?
    return get_files(config.export_path)


PlaceId = str
ListId = str
ListName = str


@dataclass(eq=True, frozen=True)
class Location:
    lat: float
    lon: float

    @property
    def url(self) -> str:
        return f'https://maps.google.com/?q={self.lat},{self.lon}'


@dataclass(unsafe_hash=True)
class Place:
    id: PlaceId
    list_name: ListName  # TODO maybe best to keep list id?
    created_at: datetime_aware  # TODO double check it's utc?
    updated_at: datetime_aware  # TODO double check it's utc?
    title: str
    location: Location
    address: Optional[str]
    note: Optional[str]

    @property
    def place_url(self) -> str:
        title = quote(self.title)
        return f'https://www.google.com/maps/place/{title}/data=!4m2!3m1!1s{self.id}'

    @property
    def location_url(self) -> str:
        return self.location.url


def _process_one(f: Path):
    with sqlite_connection(f, row_factory='row') as conn:
        msg: Any

        lists: dict[ListId, ListName] = {}
        for row in conn.execute('SELECT * FROM sync_item_data WHERE corpus == 13'):  # 13 looks like lists (e.g. saved/favorited etc)
            server_id = row['server_id']

            if server_id is None:
                # this is the case for Travel plans, Followed places, Offers
                # todo alternatively could use string_index column instead maybe?
                continue

            blob = row['item_proto']
            msg = parse_list(blob)
            name = msg.f1.name
            lists[server_id] = name

        for row in conn.execute('SELECT * FROM sync_item_data WHERE corpus == 11'):  # this looks like 'Labeled' list
            ts = row['timestamp'] / 1000
            created = datetime.fromtimestamp(ts, tz=timezone.utc)

            server_id = row['server_id']
            [item_type, item_id] = server_id.split(':')
            if item_type != '3':
                # the ones that are not 3 are home/work address?
                continue

            blob = row['item_proto']
            msg = parse_labeled(blob)
            address = msg.address.full
            if address == '':
                address = None

            location = Location(lat=row['latitude_e6'] / 1e6, lon=row['longitude_e6'] / 1e6)

            yield Place(
                id=item_id,
                list_name='Labeled',
                created_at=created,
                updated_at=created,  # doesn't look like it has 'updated'?
                title=msg.title,
                location=location,
                address=address,
                note=None,  # don't think these allow notes
            )

        for row in conn.execute('SELECT * FROM sync_item_data WHERE corpus == 14'):  # this looks like actual individual places
            server_id = row['server_id']
            [list_id, _, id1, id2] = server_id.split(':')
            item_id = f'{id1}:{id2}'

            list_name = lists[list_id]

            blob = row['item_proto']
            msg = parse_place(blob)
            title = msg.f1.title
            note = msg.f1.note
            if note == '':  # seems that protobuf does that?
                note = None

            # TODO double check timezone
            created = datetime.fromtimestamp(msg.f1.created.seconds, tz=timezone.utc).replace(microsecond=msg.f1.created.nanos // 1000)

            # NOTE: this one seems to be the same as row['timestamp']
            updated = datetime.fromtimestamp(msg.f1.updated.seconds, tz=timezone.utc).replace(microsecond=msg.f1.updated.nanos // 1000)

            address = msg.f2.addr1  # NOTE: there is also addr2, but they seem identical :shrug:
            if address == '':
                address = None

            location = Location(lat=row['latitude_e6'] / 1e6, lon=row['longitude_e6'] / 1e6)

            place = Place(
                id=item_id,
                list_name=list_name,
                created_at=created,
                updated_at=updated,
                title=title,
                location=location,
                address=address,
                note=note,
            )

            # ugh. in my case it's violated by one place by about 1 second??
            # assert place.created_at <= place.updated_at
            yield place


def saved() -> Iterator[Res[Place]]:
    def it() -> Iterator[Res[Place]]:
        paths = inputs()
        total = len(paths)
        width = len(str(total))
        for idx, path in enumerate(paths):
            logger.info(f'processing [{idx:>{width}}/{total:>{width}}] {path}')
            yield from _process_one(path)
    return unique_everseen(it)


# Summary of databases on Android (as of 20240101)
# -1_optimized_threads.notifications.db -- empty
# 1_optimized_threads.notifications.db  -- empty
# 1_tasks.notifications.db              -- empty
# -1_threads.notifications.db           -- empty
# 1_threads.notifications.db            -- doesn't look like anything interested, some trip anniversaries etc?
# 1_thread_surveys.notifications.db     -- empty
# 2_threads.notifications.db            -- empty
# accounts.notifications.db             -- just one row with account id
# brella_example_store                  -- empty
# gmm_myplaces.db                       -- contains just a few places? I think it's a subset of "Labeled"
# gmm_storage.db                        -- pretty huge, like 50Mb. I suspect it contains cache for places on maps or something
# gmm_sync.db                           -- processed above
# gnp_fcm_database                      -- list of accounts
# google_app_measurement_local.db       -- empty
# inbox_notifications.db                -- nothing interesting
# <email>_room_notifications.db         -- trip anniversaties?
# lighter_messaging_1.db                -- empty
# lighter_messaging_2.db                -- empty
# lighter_registration.db               -- empty
# peopleCache_<email>_com.google_14.db  -- contacts cache or something
# portable_geller_<email>.db            -- looks like analytics
# primes_example_store                  -- looks like analytics
# pseudonymous_room_notifications.db    -- looks like analytics
# ue3.db                                -- empty
# ugc_photos_location_data.db           -- empty
# ugc-sync.db                           -- empty
# updates-tab-visit.db                  -- empty



from my.tests.common import skip_if_not_karlicoss as pytestmark

from datetime import date, time

# todo private test.. move away
def test_tz() -> None:
    from my.jawbone import sleeps_by_date  # type: ignore[attr-defined]
    sleeps = sleeps_by_date()
    for s in sleeps.values():
        assert s.sleep_start.tzinfo is not None
        assert s.sleep_end.tzinfo is not None

    for dd, exp in [
            (date(year=2015, month=8 , day=28), time(hour=7, minute=20)),
            (date(year=2015, month=9 , day=15), time(hour=6, minute=10)),
    ]:
        sleep = sleeps[dd]
        end = sleep.sleep_end

        assert end.time() == exp
        # TODO fuck. on 0909 I woke up at around 6 according to google timeline
        # but according to jawbone, it was on 0910?? eh. I guess it's just shitty tracking.



"""
Github data: events, comments, etc. (API data)
"""
REQUIRES = [
    'git+https://github.com/karlicoss/ghexport',
]
from dataclasses import dataclass

from my.core import Paths
from my.config import github as user_config


@dataclass
class github(user_config):
    '''
    Uses [[https://github.com/karlicoss/ghexport][ghexport]] outputs.
    '''

    export_path: Paths
    '''path[s]/glob to the exported JSON data'''

###

from my.core.cfg import make_config, Attrs
def migration(attrs: Attrs) -> Attrs:
    export_dir = 'export_dir'
    if export_dir in attrs: # legacy name
        attrs['export_path'] = attrs[export_dir]
        from my.core.warnings import high
        high(f'"{export_dir}" is deprecated! Please use "export_path" instead."')
    return attrs
config = make_config(github, migration=migration)


try:
    from ghexport import dal
except ModuleNotFoundError as e:
    from my.core.hpi_compat import pre_pip_dal_handler

    dal = pre_pip_dal_handler('ghexport', e, config, requires=REQUIRES)

############################

from functools import lru_cache
from typing import Tuple, Dict, Sequence, Optional

from my.core import get_files, Path, LazyLogger
from my.core.common import mcachew

from .common import Event, parse_dt, Results, EventIds


logger = LazyLogger(__name__)


def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


def _dal() -> dal.DAL:
    sources = inputs()
    return dal.DAL(sources)


@mcachew(depends_on=inputs)
def events() -> Results:
    from my.core.common import ensure_unique
    key = lambda e: object() if isinstance(e, Exception) else e.eid
    # crap. sometimes API events can be repeated with exactly the same payload and different id
    # yield from ensure_unique(_events(), key=key)
    yield from _events()


def _events() -> Results:
    dal = _dal()
    for d in dal.events():
        if isinstance(d, Exception):
            yield d
        try:
            yield _parse_event(d)
        except Exception as e:
            yield e


from my.core import stat, Stats
def stats() -> Stats:
    return {
        **stat(events),
    }


@lru_cache(None)
def _log_if_unhandled(e) -> None:
    logger.warning('unhandled event type: %s', e)


# TODO hmm. need some sort of abstract syntax for this...
# TODO split further, title too
Link = str
EventId = str
Body = str
def _get_summary(e) -> Tuple[str, Optional[Link], Optional[EventId], Optional[Body]]:
    # TODO would be nice to give access to raw event within timeline
    dts = e['created_at']
    eid = e['id']
    tp = e['type']
    pl = e['payload']
    rname = e['repo']['name']

    mapping = {
        'CreateEvent': 'created',
        'DeleteEvent': 'deleted',
    }

    if tp == 'ForkEvent':
        url = e['payload']['forkee']['html_url']
        return f"{rname}: forked", url, None, None
    elif tp == 'PushEvent':
        commits = pl['commits']
        messages = [c['message'] for c in commits]
        ref = pl['ref']
        body = '\n'.join(messages)
        eid = f'{tp}_{e["id"]}'
        return f'{rname}: pushed {len(commits)} commits to {ref}', None, eid, body
    elif tp == 'WatchEvent':
        return f"{rname}: watching", None, None, None
    elif tp in mapping:
        what = mapping[tp]
        rt  = pl['ref_type']
        ref = pl['ref']
        if what == 'created':
            # FIXME should handle deletion?...
            eid = EventIds.repo_created(dts=dts, name=rname, ref_type=rt, ref=ref)
        mref = '' if ref is None else ' ' + ref
        # todo link to branch? only contains weird API link though
        # TODO combine automatically instead
        return f"{rname}: {what} {rt}{mref}", None, eid, None
    elif tp == 'PullRequestEvent':
        pr = pl['pull_request']
        title = pr['title']

        link  = pr['html_url']
        body  = pr['body']
        action = pl['action']
        eid = EventIds.pr(dts=dts, action=action, url=link)
        return f"{rname}: {action} PR: {title}", link, eid, body
    elif tp == 'PullRequestReviewEvent':
        pr = pl['pull_request']
        title = pr['title']

        rev = pl['review']
        link = rev['html_url']
        body = rev['body']
        eid = f'{tp}_{rev["id"]}'
        return f'{rname}: reviewed PR: {title}', link, eid, body
    elif tp == 'PullRequestReviewCommentEvent':
        pr = pl['pull_request']
        title = pr['title']

        comment = pl['comment']
        link = comment['html_url']
        body = comment['body']
        eid = f'{tp}_{comment["id"]}'
        return f'{rname}: commented on PR: {title}', link, eid, body
    elif tp == 'CommitCommentEvent':
        comment = pl['comment']
        link = comment['html_url']
        body = comment['body']
        eid = f'{tp}_{comment["id"]}'
        cid = comment['commit_id']
        return f'{rname} commented on commit: {cid}', link, eid, body
    elif tp == 'IssuesEvent':
        action = pl['action']
        iss = pl['issue']
        link  = iss['html_url']
        body  = iss['body']
        title = iss['title']
        return f'{rname}: {action} issue: {title}', link, None, body
    elif tp == 'IssueCommentEvent':
        comment = pl['comment']
        link = comment['html_url']
        body = comment['body']
        iss = pl['issue']
        title = iss['title']
        return f'{rname}: commented on issue: {title}', link, 'issue_comment_' + link, body
    elif tp == 'ReleaseEvent':
        action = pl['action']
        rel = pl['release']
        tag = rel['tag_name']
        link = rel['html_url']
        body = rel['body']
        return f"{rname}: {action} [{tag}]", link, None, body
    else:
        _log_if_unhandled(tp)
        return tp, None, None, None


def _parse_event(d: Dict) -> Event:
    summary, link, eid, body = _get_summary(d)
    if eid is None:
        eid = d['id']  # meh
    return Event(
        dt=parse_dt(d['created_at']),
        summary=summary,
        link=link,
        eid=eid,
        body=body,
    )


# TODO mm. ok, not much point in deserializing as github.Event as it's basically a fancy dict wrapper?
# from github.Event import Event as GEvent # type: ignore
# # see https://github.com/PyGithub/PyGithub/blob/master/github/GithubObject.py::GithubObject.__init__
# e = GEvent(None, None, raw_event, True)



from my.tests.common import skip_if_not_karlicoss as pytestmark

# todo maybe belongs to common
from more_itertools import ilen


def test() -> None:
    from my.lastfm import scrobbles
    assert ilen(scrobbles()) > 1000


def test_datetime_ascending() -> None:
    from my.lastfm import scrobbles
    from more_itertools import pairwise
    for a, b in pairwise(scrobbles()):
        assert a.dt <= b.dt



#!/usr/bin/env python3
# TODO this should be in dashboard
from pathlib import Path
# from kython.plotting import *
from csv import DictReader
from itertools import islice

from typing import Dict, Any, NamedTuple

# sleep = []
# with open('2017.csv', 'r') as fo:
#     reader = DictReader(fo)
#     for line in islice(reader, 0, 10):
#         sleep
#         print(line)

import matplotlib.pyplot as plt # type: ignore
from numpy import genfromtxt
import matplotlib.pylab as pylab # type: ignore

pylab.rcParams['figure.figsize'] = (32.0, 24.0)
pylab.rcParams['font.size'] = 10

jawboneDataFeatures = Path(__file__).parent / 'features.csv' # Data File Path
featureDesc: Dict[str, str] = {}
for x in genfromtxt(jawboneDataFeatures, dtype='unicode', delimiter=','):
    featureDesc[x[0]] = x[1]

def _safe_float(s: str):
    if len(s) == 0:
        return None
    return float(s)

def _safe_int(s: str):
    if len(s) == 0:
        return None
    return int(float(s)) # TODO meh

def _safe_mins(s: float):
    if s is None:
        return None
    return s / 60

class SleepData(NamedTuple):
    date: str
    asleep_time: float
    awake_time: float
    total: float
    awake: float # 'awake for' from app, time awake duing sleep (seconds)
    awakenings: int
    light: float # 'light sleep' from app (seconds)
    deep: float  # 'deep sleep' from app (sec)
    quality: float # ???

    @classmethod
    def from_jawbone_dict(cls, d: Dict[str, Any]):
        return cls(
            date=d['DATE'],
            asleep_time=_safe_mins(_safe_float(d['s_asleep_time'])),
            awake_time=_safe_mins(_safe_float(d['s_awake_time'])),
            total=_safe_mins(_safe_float(d['s_duration'])),
            light=_safe_mins(_safe_float(d['s_light'])),
            deep =_safe_mins(_safe_float(d['s_deep'])),
            awake=_safe_mins(_safe_float(d['s_awake'])),
            awakenings=_safe_int(d['s_awakenings']),
            quality=_safe_float(d['s_quality']),
        )

    def is_bad(self):
        return self.deep is None and self.light is None

    # @property
    # def total(self) -> float:
    #     return self.light + self.deep



def iter_useful(data_file: str):
    with open(data_file) as fo:
        reader = DictReader(fo)
        for d in reader:
            dt = SleepData.from_jawbone_dict(d)
            if not dt.is_bad():
                yield dt

# TODO <<< hmm. these files do contain deep and light sleep??
# also steps stats??
from my.config import jawbone as config  # type: ignore[attr-defined]

p = config.export_dir / 'old_csv'
# TODO with_my?
files = [
    p / "2015.csv",
    p / "2016.csv",
    p / "2017.csv",
]

from kython import concat, parse_date  # type: ignore
useful = concat(*(list(iter_useful(str(f))) for f in files))

# for u in useful:
#     print(f"{u.total} {u.asleep_time} {u.awake_time}")
#     # pprint(u.total)
#     pprint(u)
#     pprint("---")

dates = [parse_date(u.date, yearfirst=True, dayfirst=False) for u in useful]
# TODO filter outliers?

# TODO don't need this anymore? it's gonna be in dashboards package
from kython.plotting import plot_timestamped  # type: ignore
for attr, lims, mavg, fig in [
        ('light', (0, 400), 5, None),
        ('deep', (0, 600), 5, None),
        ('total', (200, 600), 5, None),
        ('awake_time', (0, 1200), None, 1),
        ('asleep_time', (-100, 1000), None, 1),
        # ('awakenings', (0, 5)),
]:
    dates_wkd = [d for d in dates if d.weekday() < 5]
    dates_wke = [d for d in dates if d.weekday() >= 5]
    for dts, dn in [
            (dates, 'total'),
            (dates_wkd, 'weekday'),
            (dates_wke, 'weekend')
    ]:
        mavgs = []
        if mavg is not None:
            mavgs.append((mavg, 'green'))
        fig = plot_timestamped(
            dts,
            [getattr(u, attr) for u in useful],
            marker='.',
            ratio=(16, 4),
            mavgs=mavgs,
            ylimits=lims,
            ytick_size=60,
            # figure=1,
           )
        plt.savefig(f'{attr}_{dn}.png')

# TODO use proper names?
# plt.savefig('res.png')
# fig.show()



"""
Tinder data from Android app database (in =/data/data/com.tinder/databases/tinder-3.db=)
"""
from __future__ import annotations

from collections import defaultdict, Counter
from dataclasses import dataclass
from datetime import datetime, timezone
from itertools import chain
from pathlib import Path
import sqlite3
from typing import Sequence, Iterator, Union, Dict, List, Mapping

from my.core import Paths, get_files, Res, assert_never, stat, Stats, datetime_aware, make_logger
from my.core.common import unique_everseen
from my.core.error import echain
from my.core.sqlite import sqlite_connection
import my.config


logger = make_logger(__name__)


@dataclass
class config(my.config.tinder.android):
    # paths[s]/glob to the exported sqlite databases
    export_path: Paths


@dataclass(unsafe_hash=True)
class Person:
    id: str
    name: str
    # todo bio? it might change, not sure what do we want here


@dataclass(unsafe_hash=True)
class _BaseMatch:
    # for android, checked directly shortly after a match
    when: datetime_aware
    id: str


@dataclass(unsafe_hash=True)
class _Match(_BaseMatch):
    person_id: str


@dataclass(unsafe_hash=True)
class Match(_BaseMatch):
    person: Person


# todo again, not sure what's the 'optimal' field order? perhaps the one which gives the most natural sort?
# so either match id or datetime
@dataclass
class _BaseMessage:
    # looks like gdpr takeout does contain GMT (compared against google maps data)
    sent: datetime_aware
    id: str
    text: str


@dataclass(unsafe_hash=True)
class _Message(_BaseMessage):
    match_id: str
    from_id: str
    to_id: str


@dataclass
class Message(_BaseMessage):
    match: Match
    from_: Person
    to: Person


# todo hmm I have a suspicion it might be cumulative?
# although still possible that the user might remove/install app back, so need to keep that in mind
def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


_Entity = Union[Person, _Match, _Message]
Entity = Union[Person, Match, Message]


def _entities() -> Iterator[Res[_Entity]]:
    paths = inputs()
    total = len(paths)
    width = len(str(total))
    for idx, path in enumerate(paths):
        logger.info(f'processing [{idx:>{width}}/{total:>{width}}] {path}')
        with sqlite_connection(path, immutable=True, row_factory='row') as db:
            try:
                yield from _handle_db(db)
            except Exception as e:
                yield e


def _handle_db(db: sqlite3.Connection) -> Iterator[Res[_Entity]]:
    # profile_user_view contains our own user id
    user_profile_rows = list(db.execute('SELECT * FROM profile_user_view'))

    if len(user_profile_rows) == 0:
        # shit, sometime in 2023 profile_user_view stoppped containing user profile..
        # presumably the most common from_id/to_id would be our own username
        counter = Counter([id_ for (id_,) in db.execute('SELECT from_id FROM message UNION ALL SELECT to_id FROM message')])
        if len(counter) > 0:  # this might happen if db is empty (e.g. user got logged out)
            [(you_id, _)] = counter.most_common(1)
            yield Person(id=you_id, name='you')

    for row in chain(
        user_profile_rows,
        db.execute('SELECT * FROM match_person'),
    ):
        try:
            yield _parse_person(row)
        except Exception as e:
            # todo attach error context?
            yield e

    for row in db.execute('SELECT * FROM match'):
        try:
            yield _parse_match(row)
        except Exception as e:
            yield e

    for row in db.execute('SELECT * FROM message'):
        try:
            yield _parse_msg(row)
        except Exception as e:
            yield e


def _parse_person(row: sqlite3.Row) -> Person:
    return Person(
        id=row['id'],
        name=row['name'],
    )


def _parse_match(row: sqlite3.Row) -> _Match:
    return _Match(
        id=row['id'],
        person_id=row['person_id'],
        when=datetime.fromtimestamp(row['creation_date'] / 1000, tz=timezone.utc),
    )


def _parse_msg(row: sqlite3.Row) -> _Message:
    # note it also has raw_message_data -- not sure which is best to use..
    sent = row['sent_date']
    return _Message(
        sent=datetime.fromtimestamp(sent / 1000, tz=timezone.utc),
        id=row['id'],
        text=row['text'],
        match_id=row['match_id'],
        from_id=row['from_id'],
        to_id=row['to_id'],
    )


# todo maybe it's rich_entities method?
def entities() -> Iterator[Res[Entity]]:
    id2person: Dict[str, Person] = {}
    id2match: Dict[str, Match] = {}
    for x in unique_everseen(_entities):
        if isinstance(x, Exception):
            yield x
            continue
        if isinstance(x, Person):
            id2person[x.id] = x
            yield x
            continue
        if isinstance(x, _Match):
            try:
                person = id2person[x.person_id]
            except Exception as e:
                yield e
                continue
            m = Match(
                id=x.id,
                when=x.when,
                person=person,
            )
            id2match[x.id] = m
            yield m
            continue
        if isinstance(x, _Message):
            try:
                match = id2match[x.match_id]
                from_ = id2person[x.from_id]
                to = id2person[x.to_id]
            except Exception as e:
                yield echain(RuntimeError(f'while processing {x}'), e)
                continue
            yield Message(
                sent=x.sent,
                match=match,
                id=x.id,
                text=x.text,
                from_=from_,
                to=to,
            )
            continue
        assert_never(x)


def messages() -> Iterator[Res[Message]]:
    for x in entities():
        if isinstance(x, (Exception, Message)):
            yield x
            continue


# todo not sure, maybe it's not fundamental enough to keep here...
def match2messages() -> Iterator[Res[Mapping[Match, Sequence[Message]]]]:
    res: Dict[Match, List[Message]] = defaultdict(list)
    for x in entities():
        if isinstance(x, Exception):
            yield x
            continue
        if isinstance(x, Match):
            # match might happen without messages so makes sense to handle here
            res[x]  # just trigger creation
            continue
        if isinstance(x, Message):
            try:
                ml = res[x.match]
            except Exception as e:
                yield e
                continue
            ml.append(x)
            continue
    yield res


# TODO maybe a more natural return type is Iterator[Res[Tuple[Key, Value]]]
# but this doesn't work straight away because the key might have no corresponding values


def stats() -> Stats:
    return stat(messages)



from datetime import datetime
import json
from pathlib import Path
from typing import NamedTuple, Iterator
import warnings

from ..denylist import DenyList


class IP(NamedTuple):
    addr: str
    dt: datetime


def data() -> Iterator[IP]:
    # random IP addresses
    yield IP(addr="67.98.113.0", dt=datetime(2020, 1, 1))
    yield IP(addr="59.40.113.87", dt=datetime(2020, 2, 1))
    yield IP(addr="161.235.192.228", dt=datetime(2020, 3, 1))
    yield IP(addr="165.243.139.87", dt=datetime(2020, 4, 1))
    yield IP(addr="69.69.141.154", dt=datetime(2020, 5, 1))
    yield IP(addr="50.72.224.80", dt=datetime(2020, 6, 1))
    yield IP(addr="221.67.89.168", dt=datetime(2020, 7, 1))
    yield IP(addr="177.113.119.251", dt=datetime(2020, 8, 1))
    yield IP(addr="93.200.246.215", dt=datetime(2020, 9, 1))
    yield IP(addr="127.105.171.61", dt=datetime(2020, 10, 1))


def test_denylist(tmp_path: Path) -> None:
    tf = (tmp_path / "denylist.json").absolute()
    with warnings.catch_warnings(record=True):
        # create empty denylist (though file does not have to exist for denylist to work)
        tf.write_text("[]")

        d = DenyList(tf)

        d.load()
        assert dict(d._deny_map) == {}
        assert d._deny_raw_list == []

        assert list(d.filter(data())) == list(data())
        # no data in denylist yet
        assert len(d._deny_map) == 0
        assert len(d._deny_raw_list) == 0

        # add some data
        d.deny(key="addr", value="67.98.113.0")
        # write and reload to update _deny_map, _deny_raw_list
        d.write()
        d.load()

        assert len(d._deny_map) == 1
        assert len(d._deny_raw_list) == 1

        assert d._deny_raw_list == [{"addr": "67.98.113.0"}]

        filtered = list(d.filter(data()))
        assert len(filtered) == 9
        assert "67.98.113.0" not in [i.addr for i in filtered]

        assert dict(d._deny_map) == {"addr": {"67.98.113.0"}}

        denied = list(d.filter(data(), invert=True))
        assert len(denied) == 1

        assert denied[0] == IP(addr="67.98.113.0", dt=datetime(2020, 1, 1))

        # add some non-JSON primitive data

        d.deny(key="dt", value=datetime(2020, 2, 1))

        # test internal behavior, _deny_raw_list should have been updated,
        # but _deny_map doesn't get updated by a call to .deny
        #
        # if we change this just update the test, is just here to ensure
        # this is the behaviour

        assert len(d._deny_map) == 1

        # write and load to update _deny_map
        d.write()
        d.load()

        assert len(d._deny_map) == 2
        assert len(d._deny_raw_list) == 2

        assert d._deny_raw_list[-1] == {"dt": "2020-02-01T00:00:00"}

        filtered = list(d.filter(data()))
        assert len(filtered) == 8

        assert "59.40.113.87" not in [i.addr for i in filtered]

        with open(tf, "r") as f:
            data_json = json.loads(f.read())

        assert data_json == [
            {
                "addr": "67.98.113.0",
            },
            {
                "dt": "2020-02-01T00:00:00",
            },
        ]



# hmm, sadly pytest --import-mode importlib --pyargs my.core.tests doesn't work properly without __init__.py
# although it works if you run either my.core or my.core.tests.sqlite (for example) directly
# so if it gets in the way could get rid of this later?



import importhook

@importhook.on_import('my.twitter.gdpr')
def on_import(gdpr):
    print("EXECUTING IMPORT HOOK!")
    tweets_orig = gdpr.tweets
    def tweets_patched():
        return [t.replace('gdpr', 'GDPR') for t in tweets_orig()]
    gdpr.tweets = tweets_patched



import my.config as config

from .core import __NOT_HPI_MODULE__

from .core import warnings as W

# still used in Promnesia, maybe in dashboard?
W.high("DEPRECATED! Please import my.config directly instead.")



'''
Combined exercise data
'''
from ...core.pandas import DataFrameT, check_dataframe


@check_dataframe
def dataframe() -> DataFrameT:
    # this should be somehow more flexible...
    from ...endomondo import dataframe as EDF
    from ...runnerup  import dataframe as RDF

    import pandas as pd
    return pd.concat([
        EDF(),
        RDF(),
    ])



"""
Just an helper module for testing HPI overlays
In particular the behaviour of import_original_module function

The idea of testing is that overlays extend this module, and add their own
items to items(), and the checker asserts all overlays have contributed.
"""
from my.core import __NOT_HPI_MODULE__

from dataclasses import dataclass
from datetime import datetime
import sys
from typing import Iterator, List

NOW = datetime.now()


@dataclass
class Item:
    dt: datetime
    message: str
    path: List[str]


def get_pkg_path() -> List[str]:
    pkg = sys.modules[__package__]
    return list(pkg.__path__)


# NOTE: since we're hacking path for my.util
# imports from my. should work as expected
# (even though my.config is in the private config)
from my.config import demo

assert demo.username == 'todo'

# however, this won't work while the module is imported
# from my.util import extra
# assert extra.message == 'EXTRA'
# but it will work when we actually call the function (see below)


def items() -> Iterator[Item]:
    from my.config import demo

    assert demo.username == 'todo'

    # here the import works as expected, since by the time the function is called,
    # all overlays were already processed and paths/sys.modules restored
    from my.util import extra  # type: ignore[attr-defined]

    assert extra.message == 'EXTRA'

    yield Item(dt=NOW, message='hpi main', path=get_pkg_path())



"""
Hackernews data via Dogsheep [[hacker-news-to-sqlite][https://github.com/dogsheep/hacker-news-to-sqlite]]
"""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Iterator, Sequence, Optional

from my.core import get_files, Paths, Res, datetime_aware
from my.core.sqlite import sqlite_connection
import my.config

from .common import hackernews_link


@dataclass
class config(my.config.hackernews.dogsheep):
    # paths[s]/glob to the dogsheep database
    export_path: Paths


# todo so much boilerplate... really need some common wildcard imports?...
# at least for stuff which realistically is used in each module like get_files/Sequence/Paths/dataclass/Iterator/Optional
def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


# TODO not sure if worth splitting into Comment and Story?
@dataclass(unsafe_hash=True)
class Item:
    id: str
    type: str
    created: datetime_aware  # checked and it's utc
    title: Optional[str]  # only present for Story
    text_html: Optional[str]  # should be present for Comment and might for Story
    url: Optional[str]  # might be present for Story
    # todo process 'deleted'? fields?
    # todo process 'parent'?

    @property
    def permalink(self) -> str:
        return hackernews_link(self.id)


# TODO hmm kinda annoying that permalink isn't getting serialized
# maybe won't be such a big problem if we used hpi query directly on objects, without jsons?
# so we could just take .permalink thing


def items() -> Iterator[Res[Item]]:
    f = max(inputs())
    with sqlite_connection(f, immutable=True, row_factory='row') as conn:
        for r in conn.execute('SELECT * FROM items ORDER BY time'):
            yield Item(
                id=r['id'],
                type=r['type'],
                created=datetime.fromtimestamp(r['time'], tz=timezone.utc),
                title=r['title'],
                # todo hmm maybe a method to strip off html tags would be nice
                text_html=r['text'],
                url=r['url'],
            )



"""
[[https://hypothes.is][Hypothes.is]] highlights and annotations
"""
REQUIRES = [
    'git+https://github.com/karlicoss/hypexport',
]
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, Sequence, TYPE_CHECKING

from my.core import (
    get_files,
    stat,
    Paths,
    Res,
    Stats,
)
from my.core.cfg import make_config
from my.core.hpi_compat import always_supports_sequence
import my.config


@dataclass
class hypothesis(my.config.hypothesis):
    '''
    Uses [[https://github.com/karlicoss/hypexport][hypexport]] outputs
    '''

    # paths[s]/glob to the exported JSON data
    export_path: Paths


config = make_config(hypothesis)


try:
    from hypexport import dal
except ModuleNotFoundError as e:
    from my.core.hpi_compat import pre_pip_dal_handler

    dal = pre_pip_dal_handler('hypexport', e, config, requires=REQUIRES)


Highlight = dal.Highlight
Page = dal.Page


def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


def _dal() -> dal.DAL:
    return dal.DAL(inputs())


# TODO they are in reverse chronological order...
def highlights() -> Iterator[Res[Highlight]]:
    return always_supports_sequence(_dal().highlights())


def pages() -> Iterator[Res[Page]]:
    return always_supports_sequence(_dal().pages())


def stats() -> Stats:
    return {
        **stat(highlights),
        **stat(pages),
    }


if not TYPE_CHECKING:
    # "deprecate" by hiding from mypy
    get_highlights = highlights
    get_pages = pages



'''
A helper to make warnings a bit more visible.
TODO ideally would be great to replace with some existing solution, or find a better way,
since who looks at the terminal output?
E.g. would be nice to propagate the warnings in the UI (it's even a subclass of Exception!)
'''

import sys
from typing import Optional
import warnings

import click


def _colorize(x: str, color: Optional[str]=None) -> str:
    if color is None:
        return x

    if not sys.stderr.isatty():
        return x
    # click handles importing/initializing colorama if necessary
    # on windows it installs it if necessary
    # on linux/mac, it manually handles ANSI without needing termcolor
    return click.style(x, fg=color)


def _warn(message: str, *args, color: Optional[str]=None, **kwargs) -> None:
    stacklevel = kwargs.get('stacklevel', 1)
    kwargs['stacklevel'] = stacklevel + 2 # +1 for this function, +1 for medium/high wrapper
    warnings.warn(_colorize(message, color=color), *args, **kwargs)


def low(message: str, *args, **kwargs) -> None:
    # kwargs['color'] = 'grey' # eh, grey is way too pale
    _warn(message, *args, **kwargs)


def medium(message: str, *args, **kwargs) -> None:
    kwargs['color'] = 'yellow'
    _warn(message, *args, **kwargs)


def high(message: str, *args, **kwargs) -> None:
    '''
    Meant for deprecations, i.e. things that better get some user attention
    '''
    kwargs['color'] = 'red'
    _warn(message, *args, **kwargs)


# NOTE: deprecated -- legacy import
from warnings import warn


'''
[[https://github.com/nomeata/arbtt#arbtt-the-automatic-rule-based-time-tracker][Arbtt]] time tracking
'''

REQUIRES = ['ijson', 'cffi']
# NOTE likely also needs libyajl2 from apt or elsewhere?


from pathlib import Path
from typing import Sequence, Iterable, List, Optional


def inputs() -> Sequence[Path]:
    try:
        from my.config import arbtt as user_config
    except ImportError:
        from .core.warnings import low
        low("Couldn't find 'arbtt' config section, falling back to the default capture.log (usually in HOME dir). Add 'arbtt' section with logfiles = '' to suppress this warning.")
        return []
    else:
        from .core import get_files
        return get_files(user_config.logfiles)


from .core import dataclass, Json, PathIsh, datetime_aware
from .core.compat import fromisoformat


@dataclass
class Entry:
    '''
    For the format reference, see
    https://github.com/nomeata/arbtt/blob/e120ad20b9b8e753fbeb02041720b7b5b271ab20/src/DumpFormat.hs#L39-L46
    '''

    json: Json
    # inactive time -- in ms

    @property
    def dt(self) -> datetime_aware:
        # contains utc already
        # TODO after python>=3.11, could just use fromisoformat
        ds = self.json['date']
        elen = 27
        lds = len(ds)
        if lds < elen:
            # ugh. sometimes contains less that 6 decimal points
            ds = ds[:-1] + '0' * (elen - lds) + 'Z'
        elif lds > elen:
            # and sometimes more...
            ds = ds[:elen - 1] + 'Z'

        return fromisoformat(ds)

    @property
    def active(self) -> Optional[str]:
        # NOTE: WIP, might change this in the future...
        ait = (w for w in self.json['windows'] if w['active'])
        a = next(ait, None)
        if a is None:
            return None
        a2 = next(ait, None)
        assert a2 is None, a2 # hopefully only one can be active in a time?

        p = a['program']
        t = a['title']
        # todo perhaps best to keep it structured, e.g. for influx
        return f'{p}: {t}'


# todo multiple threads? not sure if would help much... (+ need to find offset somehow?)
def entries() -> Iterable[Entry]:
    inps = list(inputs())

    base: List[PathIsh] = ['arbtt-dump', '--format=json']

    cmds: List[List[PathIsh]]
    if len(inps) == 0:
        cmds = [base] # rely on default
    else:
        # otherwise, 'merge' them
        cmds = [base + ['--logfile', f] for f in inps]

    import ijson.backends.yajl2_cffi as ijson # type: ignore
    from subprocess import Popen, PIPE
    for cmd in cmds:
        with Popen(cmd, stdout=PIPE) as p:
            out = p.stdout; assert out is not None
            for json in ijson.items(out, 'item'):
                yield Entry(json=json)


def fill_influxdb() -> None:
    from .core.influxdb import magic_fill
    from .core.freezer import Freezer
    freezer = Freezer(Entry)
    fit = (freezer.freeze(e) for e in entries())
    # TODO crap, influxdb doesn't like None https://github.com/influxdata/influxdb/issues/7722
    # wonder if can check it statically/warn?
    fit = (f for f in fit if f.active is not None)

    # todo could tag with computer name or something...
    # todo should probably also tag with 'program'?
    magic_fill(fit, name=f'{entries.__module__}:{entries.__name__}')


from .core import stat, Stats
def stats() -> Stats:
    return stat(entries)



"""
[[https://roamresearch.com][Roam]] data
"""
from datetime import datetime, timezone
from pathlib import Path
from itertools import chain
import re
from typing import NamedTuple, Iterator, List, Optional

from .core import get_files, LazyLogger, Json

from my.config import roamresearch as config

logger = LazyLogger(__name__)


def last() -> Path:
    return max(get_files(config.export_path))


class Keys:
    CREATED    = 'create-time'
    EDITED     = 'edit-time'
    EDIT_EMAIL = 'edit-email'
    STRING     = 'string'
    CHILDREN   = 'children'
    TITLE      = 'title'
    UID        = 'uid'


class Node(NamedTuple):
    raw: Json

    # TODO not sure if UTC
    @property
    def created(self) -> datetime:
        ct = self.raw.get(Keys.CREATED)
        if ct is not None:
            return datetime.fromtimestamp(ct / 1000, tz=timezone.utc)
        # ugh. daily notes don't have create time for some reason???

        title = self.title
        if title is None:
            return self.edited # fallback TODO log?
        # the format is 'February 8th, 2020'. Fucking hell.
        m = re.fullmatch(r'(\w+) (\d+)\w+, (\d+)', title)
        if m is None:
            return self.edited # fallback TODO log?
        # strip off 'th'/'rd' crap
        dts = m.group(1) + ' ' + m.group(2) + ' ' + m.group(3)
        dt = datetime.strptime(dts, '%B %d %Y').replace(tzinfo=timezone.utc)
        return dt

    @property
    def edited(self) -> datetime:
        rt = self.raw[Keys.EDITED]
        return datetime.fromtimestamp(rt / 1000, tz=timezone.utc)

    @property
    def title(self) -> Optional[str]:
        return self.raw.get(Keys.TITLE)

    @property
    def body(self) -> Optional[str]:
        return self.raw.get(Keys.STRING)

    @property
    def children(self) -> List['Node']:
        # TODO cache? needs a key argument (because of Json)
        ch = self.raw.get(Keys.CHILDREN, [])
        return list(map(Node, ch))

    @property
    def path(self) -> str:
        username = config.username # sadly, Roam research export doesn't provide it
        return f'{username}/page/{self.uid}'

    @property
    def permalink(self) -> str:
        return f'https://roamresearch.com/#/app/{self.path}'

    @property
    def uid(self) -> str:
        u = self.raw.get(Keys.UID)
        if u is not None:
            return u
        # ugh. so None apparently means "Daily note"

        # yes, it is using US date format...
        return self.created.strftime('%m-%d-%Y')

    def empty(self) -> bool:
        # sometimes nodes are empty. two cases:
        # - no heading -- child notes, like accidental enter presses I guess
        # - heading    -- notes that haven't been created yet
        return len(self.body or '') == 0 and len(self.children) == 0

    def traverse(self) -> Iterator['Node']:
        # not sure about __iter__, because might be a bit unintuitive that it's recursive..
        yield self
        for c in self.children:
            yield from c.traverse()

    def _render(self) -> Iterator[str]:
        ss = f'[{self.created:%Y-%m-%d %H:%M}] {self.title or " "}'
        body = self.body
        sc = chain.from_iterable(c._render() for c in self.children)

        yield ss
        if body is not None:
            yield body
        yield self.permalink
        for c in sc:
            yield '|  ' + c

    def render(self) -> str:
        return '\n'.join(self._render())

    def __repr__(self):
        return f'Node(created={self.created}, title={self.title}, body={self.body})'

    @staticmethod
    def make(raw: Json) -> Iterator['Node']:
        is_empty = set(raw.keys()) == {Keys.EDITED, Keys.EDIT_EMAIL, Keys.TITLE}
        # not sure about that... but daily notes end up like that
        if is_empty:
            # todo log?
            return
        yield Node(raw)


class Roam:
    def __init__(self, raw: List[Json]) -> None:
        self.raw = raw

    @property
    def notes(self) -> List[Node]:
        return list(chain.from_iterable(map(Node.make, self.raw)))

    def traverse(self) -> Iterator[Node]:
        for n in self.notes:
            yield from n.traverse()


def roam() -> Roam:
    import json
    raw = json.loads(last().read_text())
    roam = Roam(raw)
    return roam


def print_all_notes():
    # just a demo method
    # TODO demonstrate dumping as org-mode??
    for n in roam().notes:
        print(n.render())

# TODO could generate org-mode mirror in a single file for a demo?



"""
Twitter data from official app for Android
"""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
import re
from struct import unpack_from
from typing import Iterator, Sequence

from my.core import datetime_aware, get_files, LazyLogger, Paths, Res
from my.core.common import unique_everseen
from my.core.sqlite import sqlite_connect_immutable

import my.config

from .common import permalink

logger = LazyLogger(__name__)


@dataclass
class config(my.config.twitter.android):
    # paths[s]/glob to the exported sqlite databases
    export_path: Paths


def inputs() -> Sequence[Path]:
    # NOTE: individual databases are very patchy.
    # e.g. some contain hundreds of my bookmarks, whereas other contain just a few
    # good motivation for synthetic exports
    return get_files(config.export_path)


@dataclass(unsafe_hash=True)
class Tweet:
    id_str: str
    created_at: datetime_aware
    screen_name: str
    text: str

    @property
    def permalink(self) -> str:
        return permalink(screen_name=self.screen_name, id=self.id_str)


def _parse_content(data: bytes) -> str:
    pos = 0

    def skip(count: int) -> None:
        nonlocal pos
        pos += count

    def getstring(slen: int) -> str:
        if slen == 1:
            lfmt = '>B'
        elif slen == 2:
            lfmt = '>H'
        else:
            raise RuntimeError

        (sz,) = unpack_from(lfmt, data, offset=pos)
        skip(slen)
        assert sz > 0
        assert sz <= 10000  # sanity check?

        # soo, this is how it should ideally work:
        # (ss,) = unpack_from(f'{sz}s', data, offset=pos)
        # skip(sz)
        # however sometimes there is a discrepancy between string length in header and actual length (if you stare at the data)
        # example is 1725868458246570412
        # wtf??? (see logging below)

        # ughhhh
        seps = [
            b'I\x08',
            b'I\x09',
        ]
        sep_idxs = [data[pos:].find(sep) for sep in seps]
        sep_idxs = [i for i in sep_idxs if i != -1]
        assert len(sep_idxs) > 0
        sep_idx = min(sep_idxs)

        # print("EXPECTED LEN", sz, "GOT", sep_idx, "DIFF", sep_idx - sz)

        zz = data[pos : pos + sep_idx]
        skip(sep_idx)
        return zz.decode('utf8')

    skip(2)  # always starts with 4a03?

    (xx,) = unpack_from('B', data, offset=pos)
    skip(1)
    # print("TYPE:", xx)

    # wtf is this... maybe it's a bitmask?
    slen = {
        66 : 1,
        67 : 2,
        106: 1,
        107: 2,
    }[xx]

    text = getstring(slen=slen)

    # after the main tweet text it contains entities (e.g. shortened urls)
    # however couldn't reverse engineer the schema properly, the links are kinda all over the place

    # TODO this also contains image alt descriptions?
    # see 1665029077034565633

    extracted = []
    linksep = 0x6a
    while True:
        m = re.search(b'\x6a.http', data[pos:])
        if m is None:
            break

        qq = m.start()
        pos += qq

        while True:
            if data[pos] != linksep:
                break
            pos += 1
            (sz,) = unpack_from('B', data, offset=pos)
            pos += 1
            (ss,) = unpack_from(f'{sz}s', data, offset=pos)
            pos += sz
            extracted.append(ss)

    replacements = {}
    i = 0
    while i < len(extracted):
        if b'https://t.co/' in extracted[i]:
            key = extracted[i].decode('utf8')
            value = extracted[i + 1].decode('utf8')
            i += 2
            replacements[key] = value
        else:
            i += 1

    for k, v in replacements.items():
        text = text.replace(k, v)
    assert 'https://t.co/' not in text  # make sure we detected all links

    return text


_SELECT_OWN_TWEETS = '_SELECT_OWN_TWEETS'


def get_own_user_id(conn) -> str:
    # unclear what's the reliable way to query it, so we use multiple different ones and arbitrate
    # NOTE: 'SELECT DISTINCT ev_owner_id FROM lists' doesn't work, might include lists from other people?
    res = set()
    for q in [
        'SELECT DISTINCT list_mapping_user_id FROM list_mapping',
        'SELECT DISTINCT owner_id FROM cursors',
        'SELECT DISTINCT user_id FROM users WHERE _id == 1',
    ]:
        for (r,) in conn.execute(q):
            res.add(r)
    assert len(res) == 1, res
    return str(list(res)[0])


# NOTE:
# - it also has statuses_r_ent_content which has entities' links replaced
#   but they are still ellipsized (e.g. check 1692905005479580039)
#   so let's just uses statuses_content
# - there is also timeline_created_at, but they look like made up timestamps
#   don't think they represent bookmarking time
# - timeline_type
#   7, 8, 9: some sort of notifications or cursors, should exclude
#   17: ??? some cursors but also tweets
#   18: ??? relatively few, maybe 20 of them, also they all have timeline_is_preview=1?
#       most of them have our own id as timeline_sender?
#       I think it might actually be 'replies' tab -- also contains some retweets etc
#   26: ??? very low sort index
#   28: weird, contains lots of our own tweets, but also a bunch of unrelated..
#   29: seems like it contains the favorites!
#   30: seems like it contains the bookmarks
#   34: contains some tweets -- not sure..
#   63: contains the bulk of data
#   69: ??? just a few tweets
# - timeline_data_type
#   1 : the bulk of tweets, but also some notifications etc??
#   2 : who-to-follow/community-to-join. contains a couple of tweets, but their corresponding status_id is NULL
#   8 : who-to-follow/notfication
#   13: semantic-core/who-to-follow
#   14: cursor
#   17: trends
#   27: notification
#   31: some superhero crap
#   37: semantic-core
#   42: community-to-join
# - timeline_entity_type
#   1 : contains the bulk of data -- either tweet-*/promoted-tweet-*. However some notification-* and some just contain raw ids??
#   11: some sort of 'superhero-superhero' crap
#   13: always cursors
#   15: tweet-*/tweet:*/home-conversation-*/trends-*/and lots of other crap
#   31: always notification-*
# - timeline_data_type_group
#   0 : tweets?
#   6 : always notifications??
#   42: tweets (bulk of them)
def _process_one(f: Path, *, where: str) -> Iterator[Res[Tweet]]:
    with sqlite_connect_immutable(f) as db:
        if _SELECT_OWN_TWEETS in where:
            own_user_id = get_own_user_id(db)
            where = where.replace(_SELECT_OWN_TWEETS, own_user_id)

        for (
            tweet_id,
            user_name,
            user_username,
            created_ms,
            blob,
        ) in db.execute(
            f'''
            SELECT
            statuses_status_id,
            users_name,
            users_username,
            statuses_created,
            CAST(statuses_content AS BLOB)
            FROM timeline_view
            WHERE timeline_data_type == 1       /* the only one containing tweets (among with some other stuff) */
              AND timeline_data_type_group != 6 /* excludes notifications (some of them even have statuses_bookmarked == 1) */
              AND {where}
            ORDER BY timeline_sort_index DESC
            ''',
            # TODO not sure about timeline_sort_index for favorites
        ):
            assert blob is not None  # just in case, but should be filtered by the sql query
            yield Tweet(
                id_str=tweet_id,
                # TODO double check it's utc?
                created_at=datetime.fromtimestamp(created_ms / 1000, tz=timezone.utc),
                screen_name=user_username,
                text=_parse_content(blob),
            )


def _entities(*, where: str) -> Iterator[Res[Tweet]]:
    # TODO might need to sort by timeline_sort_index again?
    def it() -> Iterator[Res[Tweet]]:
        paths = inputs()
        total = len(paths)
        width = len(str(total))
        for idx, path in enumerate(paths):
            logger.info(f'processing [{idx:>{width}}/{total:>{width}}] {path}')
            yield from _process_one(path, where=where)

    # TODO hmm maybe unique_everseen should be a decorator?
    return unique_everseen(it)


def bookmarks() -> Iterator[Res[Tweet]]:
    # NOTE: in principle we get the bulk of bookmarks via timeline_type == 30 filter
    # however we still might miss on a few (I think the timeline_type 30 only refreshes when you enter bookmarks in the app)
    # if you bookmarked in the home feed, it might end up as status_bookmarked == 1 but not necessarily as timeline_type 30
    return _entities(where='statuses_bookmarked == 1')


def likes() -> Iterator[Res[Tweet]]:
    # NOTE: similarly to bookmarks, we could use timeline_type == 29, but it's only refreshed if we actually open likes tab
    return _entities(where='statuses_favorited == 1')


def tweets() -> Iterator[Res[Tweet]]:
    # NOTE: where timeline_type == 18 covers quite a few of our on tweets, but not everything
    # querying by our own user id seems the most exhaustive
    return _entities(where=f'timeline_sender_id == {_SELECT_OWN_TWEETS}')



import os
from pathlib import Path
import re
import sys

import pytest

V = 'HPI_TESTS_KARLICOSS'

skip_if_not_karlicoss = pytest.mark.skipif(
    V not in os.environ,
    reason=f'test only works on @karlicoss data for now. Set evn variable {V}=true to override.',
)


def reset_modules() -> None:
    '''
    A hack to 'unload' HPI modules, otherwise some modules might cache the config
    TODO: a bit crap, need a better way..
    '''
    to_unload = [m for m in sys.modules if re.match(r'my[.]?', m)]
    for m in to_unload:
        del sys.modules[m]


def testdata() -> Path:
    d = Path(__file__).absolute().parent.parent.parent / 'testdata'
    assert d.exists(), d
    return d



"""
Stackexchange data (uses [[https://stackoverflow.com/legal/gdpr/request][official GDPR export]])
"""

# TODO need to merge gdpr and stexport

### config
from my.config import stackexchange as user_config
from ..core import dataclass, PathIsh, make_config, get_files
@dataclass
class stackexchange(user_config):
    gdpr_path: PathIsh  # path to GDPR zip file
config = make_config(stackexchange)
# TODO later support unpacked zip too
###

# TODO just merge all of them and then filter?.. not sure

from ..core.common import Json
from ..core.compat import fromisoformat
from typing import NamedTuple, Iterable
from datetime import datetime
class Vote(NamedTuple):
    j: Json
    # todo ip?

    @property
    def when(self) -> datetime:
        return fromisoformat(self.j['eventTime'])

    # todo Url return type?
    @property
    def link(self) -> str:
        # vote target
        l = f"https://{self.j['siteId']}/"
        t = self.j['target']
        if t == 'Comment':
            # for comments, these work?
            # - https://meta.stackexchange.com/posts/comments/943975
            # - https://meta.stackexchange.com/questions/5436/direct-link-to-a-comment#comment943975_290757
            #                                            ^question id                         ^comment id ^answer id
            # hmm, this loads very raw comments without the rest of the page?
            # - https://meta.stackexchange.com/posts/27319/comments#comment-57475
            #
            # parentPostId is the original quesion
            #    TODO is not always present? fucking hell
            #    seems like there is no way to get a hierarchical comment link.. guess this needs to be handled in Promnesia normalisation...
            # postId is the answer
            l += f"posts/comments/{self.j['commentId']}"
        elif t == 'Post':
            # https://unix.stackexchange.com/q/14841/180307
            # https://unix.stackexchange.com/a/14871/180307
            # https://unix.stackexchange.com/a/16756/180307
            # shit. links generated by stackexchange are not hierarchical
            # on the other hand seems that it works without the last bit (/180307)
            # ok, 'a' works even for questions
            l += f"a/{self.j['postId']}"
        else:
            raise RuntimeError(f'Unexpected type {t}')
        return l

    # todo expose vote type?

import json
from ..core.error import Res
def votes() -> Iterable[Res[Vote]]:
    # TODO there is also some site specific stuff in qa/ directory.. not sure if its' more detailed
    # todo should be defensive? not sure if present when user has no votes
    path = max(get_files(config.gdpr_path))
    votes_path = path / 'analytics' /  'qa\\vote.submit.json'  # yes, it does contain a backslash...
    j = json.loads(votes_path.read_text(encoding='utf-8-sig'))  # not sure why, but this encoding seems necessary
    for r in reversed(j): # they seem to be in decreasing order by default
        # TODO implement check method that would go through all properties and emit errors?
        yield Vote(r)


from ..core import stat, Stats
def stats() -> Stats:
    return stat(votes)



'''
Simple location provider, serving as a fallback when more detailed data isn't available
'''

from dataclasses import dataclass
from datetime import datetime, time, timezone
from functools import lru_cache
from typing import Sequence, Tuple, Union, cast, List, Iterator

from my.config import location as user_config

from my.location.common import LatLon, DateIsh
from my.location.fallback.common import FallbackLocation, DateExact

@dataclass
class Config(user_config):
    home: Union[
        LatLon,         # either single, 'current' location
        Sequence[Tuple[ # or, a sequence of location history
            DateIsh,    # date when you moved to
            LatLon,     # the location
        ]]
    ]

    # default ~30km accuracy
    # this is called 'home_accuracy' since it lives on the base location.config object,
    # to differentiate it from accuracy for other providers
    home_accuracy: float = 30_000.0

    # TODO could make current Optional and somehow determine from system settings?
    @property
    def _history(self) -> Sequence[Tuple[datetime, LatLon]]:
        home1 = self.home
        # todo ugh, can't test for isnstance LatLon, it's a tuple itself
        home2: Sequence[Tuple[DateIsh, LatLon]]
        if isinstance(home1[0], tuple):
            # already a sequence
            home2 = cast(Sequence[Tuple[DateIsh, LatLon]], home1)
        else:
            # must be a pair of coordinates. also doesn't really matter which date to pick?
            loc = cast(LatLon, home1)
            home2 = [(datetime.min, loc)]

        # todo cache?
        res = []
        for x, loc in home2:
            dt: datetime
            if isinstance(x, str):
                dt = datetime.fromisoformat(x)
            elif isinstance(x, datetime):
                dt = x
            else:
                dt = datetime.combine(x, time.min)
            # todo not sure about doing it here, but makes it easier to compare..
            if dt.tzinfo is None:
                dt = dt.replace(tzinfo=timezone.utc)
            res.append((dt, loc))
        res = list(sorted(res, key=lambda p: p[0]))
        return res


from ...core.cfg import make_config
config = make_config(Config)


@lru_cache(maxsize=None)
def get_location(dt: datetime) -> LatLon:
    '''
    Interpolates the location at dt
    '''
    loc = list(estimate_location(dt))
    assert len(loc) == 1
    return loc[0].lat, loc[0].lon


# TODO: in python3.8, use functools.cached_property instead?
@lru_cache(maxsize=None)
def homes_cached() -> List[Tuple[datetime, LatLon]]:
    return list(config._history)


def estimate_location(dt: DateExact) -> Iterator[FallbackLocation]:
    from my.location.fallback.common import _datetime_timestamp
    d: float = _datetime_timestamp(dt)
    hist = list(reversed(homes_cached()))
    for pdt, (lat, lon) in hist:
        if d >= pdt.timestamp():
            yield FallbackLocation(
                lat=lat,
                lon=lon,
                accuracy=config.home_accuracy,
                dt=datetime.fromtimestamp(d, timezone.utc),
                datasource='via_home')
            return
    else:
        # I guess the most reasonable is to fallback on the first location
        lat, lon = hist[-1][1]
        yield FallbackLocation(
            lat=lat,
            lon=lon,
            accuracy=config.home_accuracy,
            dt=datetime.fromtimestamp(d, timezone.utc),
            datasource='via_home')
        return



#!/usr/bin/env python3
from subprocess import check_call, DEVNULL
from shutil import copytree, ignore_patterns
import os
from os.path import abspath
from sys import executable as python
from pathlib import Path

my_repo = Path(__file__).absolute().parent


def run() -> None:
    # uses fixed paths; worth it for the sake of demonstration
    # assumes we're in /tmp/my_demo now

    # 1. clone git@github.com:karlicoss/my.git
    copytree(
        my_repo,
        'my_repo',
        symlinks=True,
        ignore=ignore_patterns('.tox*'),  # tox dir might have broken symlinks while tests are running in parallel
    )

    # 2. prepare repositories you'd be using. For this demo we only set up Hypothesis
    tox = 'TOX' in os.environ
    if tox: # tox doesn't like --user flag
        check_call(f'{python} -m pip install        git+https://github.com/karlicoss/hypexport.git'.split())
    else:
        try:
            import hypexport
        except ModuleNotFoundError:
            check_call(f'{python} -m pip --user git+https://github.com/karlicoss/hypexport.git'.split())


    # 3. prepare some demo Hypothesis data
    hypothesis_backups = abspath('backups/hypothesis')
    Path(hypothesis_backups).mkdir(exist_ok=True, parents=True)
    check_call([
        'curl',
        'https://raw.githubusercontent.com/taniki/netrights-dashboard-mockup/master/_data/annotations.json',
        '-o', hypothesis_backups + '/annotations.json',
    ], stderr=DEVNULL)
    #

    # 4. point my.config to the Hypothesis data
    mycfg_root = abspath('my_repo')
    init_file = Path(mycfg_root) / 'my/config.py'
    init_file.write_text(init_file.read_text().replace(
        '/path/to/hypothesis/data',
        hypothesis_backups,
    ))
    #

    # 4. now we can use it!
    os.chdir(my_repo)

    check_call([python, '-c', '''
import my.hypothesis

pages = my.hypothesis.pages()

from itertools import islice
for page in islice(pages, 0, 8):
    print('URL:   ' + page.url)
    print('Title: ' + page.title)
    print('{} annotations'.format(len(page.highlights)))
    print()
'''], env={
    # this is just to prevent demo.py from using real data
    # normally, it will rely on having my.config in ~/.config/my
    'MY_CONFIG': mycfg_root,
    **os.environ,
})

# that should result in something like this:

# URL:   https://tacticaltech.org/
# Title: Tactical Technology Collective
# 1 annotations
#
# URL:   https://web.hypothes.is/blog/annotating-the-wild-west-of-information-flow/
# Title: Annotating the wild west of information flow – Hypothesis
# 1 annotations
#
# URL:   http://www.liberation.fr/futurs/2016/12/12/megafichier-beauvau-prie-de-revoir-sa-copie_1534720
# Title: «Mégafichier» : Beauvau prié de revoir sa copie
# 3 annotations
#
# URL:   https://www.wired.com/2016/12/7500-faceless-coders-paid-bitcoin-built-hedge-funds-brain/
# Title: 7,500 Faceless Coders Paid in Bitcoin Built a Hedge Fund’s Brain
# 4 annotations
#
# URL:   http://realscreen.com/2016/12/06/project-x-tough-among-sundance-17-doc-shorts/
# Title: “Project X,” “Tough” among Sundance ’17 doc shorts
# 1 annotations
#
# URL:   https://grehack.fr/2016/program
# Title: GreHack | Security conference and hacking game 2016 | Grenoble
# 1 annotations
#
# URL:   https://respectmynet.eu/
# Title: [!] Respect My Net
# 1 annotations
#
# URL:   https://www.youtube.com/watch?v=Xgp7BIBtPhk
# Title: BBC Documentaries 2016: The Joy of Data [FULL BBC SCIENCE DOCUMENTARY]
# 1 annotations

from contextlib import contextmanager
@contextmanager
def named_temp_dir(name: str):
    """
    Fixed name tmp dir
    """
    import tempfile
    td = Path(tempfile.gettempdir()) / name
    try:
        td.mkdir(exist_ok=False)
        yield td
    finally:
        import os, shutil
        skip_cleanup = 'CI' in os.environ and os.name == 'nt'
        # TODO hmm for some reason cleanup on windows causes AccessError
        if not skip_cleanup:
            shutil.rmtree(str(td))


def main():
    with named_temp_dir('my_demo') as tdir:
        os.chdir(tdir)
        run()


if __name__ == '__main__':
    main()
